//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "fMain.h"
#include "fLayers.h"
#include "fEntity.h"
#include "dstring.h"
#include <Dialogs.hpp>
#include <Classes.hpp>
#include <DXFImage.hpp>
#include <SGImage.hpp>
#include <sgConsts.hpp>
#include <DXFConv.hpp>
#ifndef sgDXFONLY
#include <DWG.hpp>
#include <HPGL2.hpp>
#include <SVG.hpp>
#include <CGM.hpp>
#endif
#ifdef SG_USEDXFEXPORT
#include <DXFExport.hpp>
#endif
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "SGImage"
#pragma resource "*.dfm"

AnsiString sNotVectorial = "Raster drawings can not be saved to DXF format";

TfmMain *fmMain;
char *FSaveToFile;
//---------------------------------------------------------------------------
__fastcall TfmMain::TfmMain(TComponent* Owner)
  : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TfmMain::FormCreate(TObject *Sender)
{
  OpenDialog->Filter = GraphicFilter(__classid(TGraphic));
  FLastAdded = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TfmMain::btnOpenClick(TObject *Sender)
{
  if (OpenDialog->Execute())
  {
    sbrCoords->Panels->Items[0]->Text = "Please wait...";
    sgCADImage->LoadFromFile(OpenDialog->FileName);
    sgCADImage->Align = alNone;
    sgCADImage->Align = alClient;
  }
  sbrCoords->Panels->Items[0]->Text = OpenDialog->FileName;
}
//---------------------------------------------------------------------------
void __fastcall TfmMain::FormMouseWheelDown(TObject *Sender,
      TShiftState Shift, TPoint &MousePos, bool &Handled)
{
  if (!sgCADImage->Empty())
  {
    MousePos = sbImage->ScreenToClient(MousePos);
    sgCADImage->ChangeScale(false, 0.8, MousePos);
    Handled = true;
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::FormMouseWheelUp(TObject *Sender,
      TShiftState Shift, TPoint &MousePos, bool &Handled)
{
  if (!sgCADImage->Empty())
  {
    MousePos = sbImage->ScreenToClient(MousePos);
    sgCADImage->ChangeScale(false, 1.2, MousePos);
    if (sgCADImage->Width < 5)
      sgCADImage->ChangeScale(False, 2, MousePos);
    Handled = true;
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnZoomInClick(TObject *Sender)
{
  sgCADImage->ChangeScale(false, 2, sgCADImage->Center());
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnZoomOutClick(TObject *Sender)
{
  sgCADImage->ChangeScale(false, 0.5, sgCADImage->Center());
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnNewClick(TObject *Sender)
{
  TsgDXFEntity *Ent = NULL;
  TsgDXFImage *Im;

  if (!sgCADImage->Empty())
    if (sgCADImage->Picture->Graphic->InheritsFrom(__classid(TsgDXFImage)))
    {
      Im = dynamic_cast<TsgDXFImage *>(sgCADImage->Picture->Graphic);
      Ent = GetEntity(Im);
      if (Ent != NULL)
      {
        Im->Converter->Sections[csEntities]->AddEntity(Ent);
        if (Im->Converter->OnCreate != NULL)
          Im->Converter->OnCreate(Ent);
        Im->Converter->Loads(Ent);
        Im->GetExtents();
        FLastAdded = Ent;
      }
      sgCADImage->Refresh();
    }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnDeleteClick(TObject *Sender)
{

  TsgDXFImage *Im;

  if ((!sgCADImage->Empty())&&
     (sgCADImage->Picture->Graphic->InheritsFrom(__classid(TsgDXFImage))))
  {
    Im = dynamic_cast<TsgDXFImage *>(sgCADImage->Picture->Graphic);
    if (FLastAdded != NULL)
    {
      Im->Converter->DeleteEntity(FLastAdded, true);
      Im->GetExtents();
      FLastAdded = NULL;
      sgCADImage->Refresh();
    }
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnSaveToDXFClick(TObject *Sender)
{
  #ifdef SG_USEDXFEXPORT
  TsgDXFExport *E;
  TMetafile *vMet;
  TsgDXFImage *Im;
  TGraphic *vGr;
  TsgFloat vOffX, vOffY, vUSize;

  if (!SaveDialog->Execute()) return;
  if (AnsiLowerCase(ExtractFileExt(SaveDialog->FileName)) != ".dxf")
     SaveDialog->FileName = SaveDialog->FileName + ".dxf";
  FSaveToFile = SaveDialog->FileName.c_str();
  vGr = dynamic_cast<TGraphic *>(sgCADImage->Picture->Graphic);
  E = new TsgDXFExport;
  try
  {

    E->IsParseWhite = true;// False
    if (!((vGr->InheritsFrom(__classid(TsgDXFImage)) || (vGr->InheritsFrom(__classid(TMetafile)))
    {
      ShowMessage(sNotVectorial);
      return;
    }
    else
    {
      if (vGr->ClassType() != __classid(TsgDXFImage))
      {
        E->Use01MM = false;
        E->Canvas->StretchDraw(Rect(0, 0, vGr->Width, vGr->Height), vGr);
        E->EndDraw();
      }
      else
      {
        Im = dynamic_cast<TsgDXFImage *>(vGr);
        vMet = Im->ExportToMetafile(vOffX, vOffY, vUSize);
        OffsetX = vOffX;
        OffsetY = vOffY;
        E->UnitSize = vUSize;
        E->LoadFromMetafile(vMet);
        #ifdef SG_DEBUG
        vMet->SaveToFile(ChangeFileExt(FSaveToFile, ".emf"));
        #endif
        vMet->Free();
      }
    }
    E->SaveToFile(FSaveToFile);
  }
  __finally
  {
    E->Free();
  }
  #endif
  #ifndef SG_USEDXFEXPORT
    MessageBox(0, "This function is available with the DXFExport VCL only.", "WARNING", MB_ICONWARNING);
  #endif
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sgCADImageMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
  TsgDXFImage *Im;
  RECT vRect;
  TFPoint vPoint;
  #ifndef sgDXFONLY
  TsgHPGLImage * vHPGLImg;
  #endif

  if (sgCADImage->Empty()) return;
  if (sgCADImage->Picture->Graphic->InheritsFrom(__classid(TsgDXFImage)))
  {
    Im = dynamic_cast<TsgDXFImage *>(sgCADImage->Picture->Graphic);
    vRect = sgCADImage->PictureRect();
    vPoint = Im->GetCADCoords(X * 1.0 / (vRect.right - vRect.left),
       1 - Y * 1.0 / (vRect.bottom - vRect.top));
    #ifndef sgDXFONLY
    if (sgCADImage->Picture->Graphic->ClassType() == __classid(TsgHPGLImage))
    {
      vHPGLImg = dynamic_cast<TsgHPGLImage *>(sgCADImage->Picture->Graphic);
      vPoint.X = vPoint.X * vHPGLImg->MMWidth / 100 / (vHPGLImg->Extents.Right - vHPGLImg->Extents.Left);
      vPoint.Y = vPoint.Y * vHPGLImg->MMHeight / 100 / Ceil(vHPGLImg->Extents.Top - vHPGLImg->Extents.Bottom);
    }
    #endif
    TVarRec v[] = {vPoint.X, vPoint.Y};
    sbrCoords->Panels->Items[1]->Text = Format("(%.3f; %.3f; 0.000)", v, 1);
  }
}
//---------------------------------------------------------------------------

#ifdef SG_USEDXFEXPORT
bool DirectExportEntity(TsgDXFExport *DXFExport, TsgDXFEntity *DXFEntity)
{
  if (DXFEntity->PaperSpace != 0)// this export made only for 'Model Space'
    return false;
  TdxfExportData Data;
  AnsiString ent_name = DXFEntity->EntName();
  AnsiString layer_name = DXFEntity->Layer->Name;
  if (layer_name == "0")
    DXFExport->CurrentLayer = NULL;
  else
    DXFExport->CurrentLayer = new TsgExpDXFLayer(layer_name);
  DXFExport->CurrentLayer->Color = ColorToDXF(DXFEntity->Layer->Pen->Color);
  memset(&Data, 0, sizeof(TdxfExportData));
  if (ent_name == "ARC") {
     TsgDXFArc *vArc = dynamic_cast<TsgDXFArc*>(DXFEntity);
     Data.Color = ColorToDXF(vArc->Pen->Pen->Color);
     Data.Point = MakeDXFFromFPoint(vArc->Point);
     Data.Radius = vArc->Radius;
     Data.StartAngle = vArc->StartAngle;
     Data.EndAngle = vArc->EndAngle;
     DXFExport->AddArc(Data);
  }
  else if (ent_name == "CIRCLE") {
     TsgDXFCircle *vCircle = dynamic_cast<TsgDXFCircle*>(DXFEntity);
     Data.Color = ColorToDXF(vCircle->Pen->Pen->Color);
     Data.Point = MakeDXFFromFPoint(vCircle->Point);
     Data.Radius = vCircle->Radius;
     DXFExport->AddCircle(Data);
  }
  else if (ent_name == "ELLIPSE") {
     TsgDXFEllipse *vEllipse = dynamic_cast<TsgDXFEllipse*>(DXFEntity);
     Data.Color = ColorToDXF(vEllipse->Pen->Pen->Color);
     Data.Point = MakeDXFFromFPoint(vEllipse->Point);
     Data.Point1 = MakeDXFFromFPoint(vEllipse->RadPt);
     Data.Radius = vEllipse->Ratio;
     Data.StartAngle = Radian(vEllipse->StartAngle);
     Data.EndAngle = Radian(vEllipse->EndAngle);
     Data.SelfType = atEllipse;
     DXFExport->AddArc(Data);
  }
  else if (StrPos(ent_name.c_str(), "INSERT") != NULL) {
    TsgDXFInsert *vInsert = dynamic_cast<TsgDXFInsert*>(DXFEntity);
    if (vInsert->Block != NULL)
    {
      Data.Text = vInsert->Block->Name;
      Data.Point = MakeDXFFromFPoint(vInsert->Point);
      Data.Scale = MakeDXFFromFPoint(vInsert->Scale);
      DXFExport->Insert(Data);
    }
  }
  else if (ent_name == "LINE") {
     TsgDXFLine *vLine = dynamic_cast<TsgDXFLine*>(DXFEntity);
     Data.Color = ColorToDXF(vLine->Pen->Color);
     Data.Thickness = vLine->Thickness;
     Data.Point = MakeDXFFromFPoint(vLine->Point);
     Data.Point1 = MakeDXFFromFPoint(vLine->Point1);
     DXFExport->AddLine(Data);
  }
  else if (StrPos(ent_name.c_str(), "MTEXT") != NULL) {
     TsgDXFMText *vMText = dynamic_cast<TsgDXFMText*>(DXFEntity);
     Data.Color = ColorToDXF(vMText->Color);
     Data.Point = MakeDXFFromFPoint(vMText->Point);
     Data.Point1 = MakeDXFFromFPoint(vMText->Point1);
     Data.FHeight = vMText->Height;
     Data.Text = vMText->Text;
     DXFExport->AddMText(Data);
  }
  else if (StrPos(ent_name.c_str(), "TEXT") != NULL) {
     TsgDXFText *vText = dynamic_cast<TsgDXFText*>(DXFEntity);
     Data.Color = ColorToDXF(vText->Font->Color);
     Data.Point = MakeDXFFromFPoint(vText->Point);
     Data.FHeight = vText->Height;
     Data.HAlign = vText->HAlign;
     Data.Rotation = vText->Rotation;
     Data.Text = vText->Text;
     DXFExport->AddText(Data);
  }
  else if ((ent_name == "POLYLINE") || (ent_name == "LWPOLYLINE")) {
    TsgDXFPolyline *vPolyline = dynamic_cast<TsgDXFPolyline*>(DXFEntity);
    TdxfPoint *pdxf_point, dxf_point;
    int i;
    Data.Color = ColorToDXF(vPolyline->Pen->Pen->Color);
    Data.Count = 1;
    Data.Points = new TList();
    Data.Points->Add(new TList());
    // Export vertexes without bulges and meshes
    for (i= 0; i< vPolyline->PolyPoints->Count; ++i)
    {
      pdxf_point = new TdxfPoint();
      TFPoint *p = (TFPoint*)(vPolyline->PolyPoints->Items[i]);
      pdxf_point->X = p->X;
      pdxf_point->Y = p->Y;
      pdxf_point->Z = p->Z;
      ((TList*) Data.Points->Items[0])->Add(pdxf_point);
    }
    DXFExport->AddPolyLine(Data, 0);
    // Lists have been copied to internal data
    for (i= 0; i< ((TList*) Data.Points->Items[0])->Count; ++i)
      delete (((TList*) Data.Points->Items[0])->Items[i]);
    delete ((TList*) Data.Points->Items[0]);
    delete Data.Points;
  }
  else if (ent_name == "SOLID") {
     TsgDXFSolid *vSolid = dynamic_cast<TsgDXFSolid*>(DXFEntity);
     Data.Color = ColorToDXF(vSolid->Pen->Color);
     Data.Point = MakeDXFFromFPoint(vSolid->Point);
     Data.Point1 = MakeDXFFromFPoint(vSolid->Point1);
     Data.Point2 = MakeDXFFromFPoint(vSolid->Point2);
     Data.Point3 = MakeDXFFromFPoint(vSolid->Point3);
     DXFExport->AddSolid(Data);
  }
  else if (ent_name == "SPLINE") {
    // Has not full realization in "DXFExport" for direct export
    // (it is possible with the big accuracy to export as Polyline)
  }
  else if (ent_name == "HATCH"){
    // Has not full realization in "DXFExport" for direct export
  }
  return true;
}
#endif

//---------------------------------------------------------------------------
void __fastcall TfmMain::btnDirectExportClick(TObject *Sender)
{
  #ifdef SG_USEDXFEXPORT
  if (!(SaveDialog->Execute())) return;
  if (AnsiLowerCase(ExtractFileExt(SaveDialog->FileName)) != ".dxf")
     SaveDialog->FileName = SaveDialog->FileName + ".dxf";

  TsgDXFImage *Img;
  TsgDXFExport *vDXF;
  TdxfExportData Data;
  TsgExpDXFLayer *vLayer;
  PdxfPoint Pt;
  int cnt, i, j;
  vDXF = new TsgDXFExport();
  try
  {
    Img = dynamic_cast<TsgDXFImage *>(sgCADImage->Picture->Graphic);
    ////////////////////
    //LTYPES SECTION//
    ////////////////////
    float *Array = new float[20];
    memset(Array, 0, sizeof(float) * 20);
    cnt = Img->Converter->Counts[csLTypes];
    for (int i = 0; i < cnt; i++){
        AnsiString LTypeName = Img->Converter->LTypes[i]->Name;
        if ((LTypeName == "Continuous") || (LTypeName == "ByLayer") || (LTypeName == "ByBlock"))
          continue;
        //AnsiString LTypePatternLength = Img->Converter->LTypes[i]->Lines->PatternLength;
        int ticks_cnt = Img->Converter->LTypes[i]->Lines->Count();
        for (int j = 0; j < ticks_cnt; j++){
            Array[j] = Img->Converter->LTypes[i]->Lines->Ticks[j];
        }
        vDXF->AddLType(LTypeName, Array, ticks_cnt);
    }
    delete[] Array;

    ////////////////////
    //BLOCKS SECTION//
    ////////////////////
    cnt = Img->Converter->Counts[csBlocks];
    for (i = 0; i < cnt; i++){
        AnsiString BlockName = Img->Converter->Blocks[i]->Name;
        int ent_cnt = Img->Converter->Blocks[i]->Count;
        if (ent_cnt < 1)
          continue;
        vDXF->BeginBlock(BlockName);
        for (j = 0; j < ent_cnt; j++) {
            DirectExportEntity(vDXF, Img->Converter->Blocks[i]->Entities[j]);
        }
        vDXF->EndBlock();
    }
    ////////////////////
    //ENTITIES SECTION//
    ////////////////////
    cnt = Img->Converter->Counts[csEntities];
    for (i = 0; i < cnt; i++){
        DirectExportEntity(vDXF, Img->Converter->Entities[i]);
    }
    vDXF->SaveToFile(SaveDialog->FileName);
  }
  __finally
  {
    delete vDXF;
  }
  #else
  Application->MessageBox("This function is available with the DXFExport VCL only.", "WARNING", MB_OK | MB_ICONWARNING);
  #endif
}

//---------------------------------------------------------------------------
TsgDXFBlock* FindBlock(TsgDXFConverter *AConverter, const AnsiString AName)
{
  TsgDXFBlock *vBlock;
  for (int i=0; i< AConverter->Counts[csBlocks]; ++i)
  {
    vBlock = AConverter->Blocks[i];
    if (vBlock->Name != AName)
      continue;
    return vBlock;
  }
  return NULL;
}

void __fastcall TfmMain::mmiExitClick(TObject *Sender)
{
  Close();  
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::mmiFitToWindowClick(TObject *Sender)
{
  sgCADImage->Align = alNone;
  sgCADImage->Align = alClient;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::mmiAboutClick(TObject *Sender)
{
  MessageBox(0, "DWG/DXF VCLs demo version \n Please contact www.cadsofttools.com to order a full version", "WARNING", MB_ICONWARNING);
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnShowLayersClick(TObject *Sender)
{
  TsgDXFImage *Im;
  if (!sgCADImage->Empty())
    if (sgCADImage->Picture->Graphic->InheritsFrom(__classid(TsgDXFImage)))
    {
      Im = dynamic_cast<TsgDXFImage *>(sgCADImage->Picture->Graphic);
      LayersDialogExecute(Im, sgCADImage);
    }
}
//---------------------------------------------------------------------------

 