/* ---------------------------------------------------------------------------
  How to get entities properties.

  1. Create procedure of type:
    TsgCADEntityProc = procedure(Entity: TsgDXFEntity) of object;

    in this demo:
    procedure TForm1.ReadCADEntities(Entity: TsgDXFEntity);

   2. Crate TsgCADIterate object:
    FCADParams: TsgCADIterate; in this demo


   3. Create TsgDXFImage object (in the demo it is made in Button1Click
   on   Image1.Picture.LoadFromFile(OpenPictureDialog1.FileName);
   and stored in Image1.Picture.Graphic.

   4.  Set (TsgDXFImage object).Converer.AutoInsert := True
   See
   TsgDXFImage(Image1.Picture.Graphic).Converter.AutoInsert := True; // to get all the elements inside of inserts

   5. Call (TsgDXFImage object).Converer.Iterate with parameters made in items 1. and 2.
     TsgDXFImage(Image1.Picture.Graphic).Converter.Iterate(ReadCADEntities, FCADParams);

   6.  in procedure made in item 1. read all the entities:
       procedure TForm1.ReadCADEntities(Entity: TsgDXFEntity);
       begin
       //  use Entity here, see documentation about properties of different classes,
       //  allows to get lines, text, arcs, etc.
       end;
*/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "Unit1.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"

TForm1 *Form1;
TFPoint PtXMat(TFPoint P, TFMatrix M);
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------

float Part(int i, TFPoint P, TFMatrix M)
{
    return P.X * M[0][i] + P.Y * M[1][i] + P.Z * M[2][i] + M[3][i];
}
//---------------------------------------------------------------------------

TFPoint PtXMat(TFPoint P, TFMatrix M)
{
  TFPoint P1;
  P1.X = Part(0, P, M);
  P1.Y = Part(1, P, M);
  P1.Z = Part(2, P, M);
  return P1;
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ReadCADEntities(TsgDXFEntity * Entity)
{
  AnsiString PS[4] = {"psSolid", "psDash", "psDot", "psDashDot"};

  AnsiString S;
  TsgDXFLayer *L;
  TColor C;

  S = "ClassName=" + Entity->ClassName();
  S = S + "; Entity name=" + Entity->EntName();
  DoScale2D(FCADParams); // calculates 2d scale and rotation
  // layer
  L = EntLayer(Entity, FCADParams.Insert);
  if (L) S = S + " layer = " + L->Name;
  // color
  S = S + " style = " + PS[(int)EntStyle(Entity)];
  C = EntColor(Entity, FCADParams.Insert);
  if (C == clNone)
    S = S + " color=black/white";
  else
    S = S + " color = " +  IntToHex((int)C, -6) + " (" + ColorToString(C) + ")";
  // particular properties
  TextFile->Add(S);

  //Face - see dxfimage.pas TsgDXFImage.DrawFace
  //Solid  - see dxfimage.pas TsgDXFImage.DrawSolid
  if (Entity->ClassType() == __classid(TsgDXFSolid)) ImportSolid(Entity);
  else
  //point - see dxfimage.pas TsgDXFImage.DrawPoint
  //line  - see dxfimage.pas TsgDXFImage.DrawLine
  if (Entity->ClassType() == __classid(TsgDXFLine)) ImportLine(Entity);
  else
  //viewport - see dxfimage.pas TsgDXFImage.Draw
  //hatch - see dxfimage.pas TsgDXFImage.DrawHatch
  //ellipse - see dxfimage.pas TsgDXFImage.DrawEllipse
  if (Entity->ClassType() == __classid(TsgDXFEllipse)) ImportEllipse(Entity);
  else
  //arc - see dxfimage.pas TsgDXFImage.DrawArc
  if (Entity->ClassType() == __classid(TsgDXFArc)) ImportArc(Entity);
  else
  //Circle - see dxfimage.pas TsgDXFImage.DrawCircle
  if (Entity->ClassType() == __classid(TsgDXFCircle)) ImportCircle(Entity);
  else
  //Leader - see dxfimage.pas TsgDXFImage.DrawLeader
  //Spline - see dxfimage.pas TsgDXFImage.DrawSpline
  if (Entity->ClassType() == __classid(TsgDXFSpline)) ImportSpline(Entity);
  else
  //PolyLine - see dxfimage.pas TsgDXFImage.DrawPolyLine
  if (Entity->ClassType() == __classid(TsgDXFPolyline)) ImportPolyLine(Entity);
  else
  //attdef - see dxfimage.pas TsgDXFImage.DrawText
  if (Entity->ClassType() == __classid(TsgDXFAttdef)) ImportAttdef(Entity);
  else
  //text - see dxfimage.pas TsgDXFImage.DrawText
  if (Entity->ClassType() == __classid(TsgDXFText)) ImportText(Entity);
}

//---------------------------------------------------------------------------

void __fastcall TForm1::BtnImportClick(TObject *Sender)
{
  AnsiString vLogFileName;
  TsgDXFConverter *Conv;

  #ifdef sgDXFONLY
  OpenPictureDialog1->FilterIndex = 2;
  #endif

  OpenPictureDialog1->FileName = "";
  TextFile = new TStringList();
  try
  {
    if (OpenPictureDialog1->Execute())
    {
      Image1->Picture->LoadFromFile(OpenPictureDialog1->FileName);
      FCurCell = 0;
      FCADParams.Matrix[0][0] = 1;
      FCADParams.Matrix[1][1] = 1;
      FCADParams.Matrix[2][2] = 1;
      Conv = ((TsgDXFImage *)(Image1->Picture->Graphic))->Converter;
      Conv->ImportMode = imImport;
      Conv->AutoInsert = true; // to get all the elements inside of inserts
      Conv->Iterate((TsgCADEntityProc)(& ReadCADEntities), FCADParams);
      SaveDialog->FileName = ChangeFileExt(OpenPictureDialog1->FileName, ".txt");
      if (SaveDialog->Execute())
        TextFile->SaveToFile(SaveDialog->FileName);
    }
  }
  __finally
  {
    delete TextFile;
  }
}
//---------------------------------------------------------------------------


void __fastcall TForm1::ImportSolid(TObject *Sender)
{
  TFPoint P;
  P = PtXMat(((TsgDXFSolid *)Sender)->Point, FCADParams.Matrix);
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format("P1: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  P = PtXMat(((TsgDXFSolid *)Sender)->Point1, FCADParams.Matrix);
  TVarRec v1[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format("P2: X=%.6f Y=%.6f Z=%.6f", v1, ARRAYSIZE(v1)-1));
  P = PtXMat(((TsgDXFSolid *)Sender)->Point2, FCADParams.Matrix);
  TVarRec v2[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format("P3: X=%.6f Y=%.6f Z=%.6f", v2, ARRAYSIZE(v2)-1));
  P = PtXMat(((TsgDXFSolid *)Sender)->Point3, FCADParams.Matrix);
  TVarRec v3[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format("P4: X=%.6f Y=%.6f Z=%.6f", v3, ARRAYSIZE(v3)-1));
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportLine(TObject *Sender)
{
  TFPoint P;
  P = PtXMat(((TsgDXFLine *)Sender)->Point, FCADParams.Matrix);
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format(" Begin point: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  P = PtXMat(((TsgDXFLine *)Sender)->Point1, FCADParams.Matrix);
  TVarRec v1[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format(" End point: X=%.6f Y=%.6f Z=%.6f", v1, ARRAYSIZE(v1)-1));
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportEllipse(TObject *Sender)
{
  TFPoint P;
  P = PtXMat(((TsgDXFEllipse *)Sender)->Point, FCADParams.Matrix);
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add("Center of Ellipse: ");
  TextFile->Add(Format(" Begin point: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  TVarRec v1[] = {((TsgDXFEllipse*)Sender)->StartAngle};
  TextFile->Add(Format(" Start Angle: %.6f", v1, ARRAYSIZE(v1)-1));
  TVarRec v2[] = {((TsgDXFEllipse*)Sender)->EndAngle};
  TextFile->Add(Format(" End Angle: %.6f", v2, ARRAYSIZE(v2)-1));
  TVarRec v3[] = {((TsgDXFEllipse *)Sender)->Radius,
    ((TsgDXFEllipse *)Sender)->Radius * ((TsgDXFEllipse *)Sender)->Ratio};
  TextFile->Add(Format(" Rx: %.6f  Ry: %.6f ", v3, ARRAYSIZE(v3)-1));
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportArc(TObject *Sender)
{
  TFPoint P;
  P = PtXMat(((TsgDXFArc *)Sender)->Point, FCADParams.Matrix);
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format(" Center of RoundArc: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  TVarRec v1[] = {((TsgDXFArc *)Sender)->StartAngle};
  TextFile->Add(Format(" Start Angle: %.6f", v1, ARRAYSIZE(v1)-1));
  TVarRec v2[] = {((TsgDXFArc *)Sender)->EndAngle};
  TextFile->Add(Format(" End Angle: %.6f", v2, ARRAYSIZE(v2)-1));
  TVarRec v3[] = {((TsgDXFArc *)Sender)->Radius};
  TextFile->Add(Format(" Rx: %.6f", v3, ARRAYSIZE(v3)-1));;
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportCircle(TObject *Sender)
{
  TFPoint P;
  P = PtXMat(((TsgDXFCircle *)Sender)->Point, FCADParams.Matrix);
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format(" Center of Circle: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportPolyLine(TObject *Sender)
{
  int I;
  TFPoint P;
  AnsiString S;
  TsgDXFVertex *Vertex;
  TsgDXFPolyline *Poly = (TsgDXFPolyline *)Sender;

  if (Poly->Closed)
    TextFile->Add(" Points of Rect:");
  else
    TextFile->Add(" Points of PolyLine:");
  int Cnt = Poly->Count;
  for(I = 0; I < Cnt; I++)
  {
    Vertex = (TsgDXFVertex *)(Poly->Entities[I]);
    P = PtXMat(Vertex->Point, FCADParams.Matrix);
    TVarRec v[] = {P.X, P.Y, P.Z};
    TextFile->Add(" P" + IntToStr(I + 1) + ": " +
      Format(" X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  }
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportAttdef(TObject *Sender)
{
  TFPoint P;
  AnsiString S1, S2;
  TsgDXFAttrib *CurAttrib;
  TsgDXFAttdef *vAttDef = (TsgDXFAttdef *)Sender;
  TsgDXFText *vText = (TsgDXFText *)Sender;

  DoScale2D(FCADParams);
  vAttDef->DoNewText(vAttDef->Tag);
  S1 = vText->Text;
  P = PtXMat(vText->StartPoint, FCADParams.Matrix);
  S2 = " Angle=" + FloatToStr(vText->Rotation);
  if (FCADParams.Insert)
  {
    CurAttrib = FCADParams.Insert->Attrib(S1, P);
    if (CurAttrib)
    {
      S1 = CurAttrib->Text;
      P = CurAttrib->Box.TopLeft;
      S2 = " Angle=" + FloatToStr(CurAttrib->Rotation);
    }
  }
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format("Start point of Attdef: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  TextFile->Add(" Text of Attdef: " + S1 + S2);
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportText(TObject *Sender)
{
  TFPoint P;
  AnsiString S1, S2;
  TsgDXFText *vText = (TsgDXFText *)Sender;

  DoScale2D(FCADParams);
  S1 = vText->Text;
  P = PtXMat(vText->StartPoint, FCADParams.Matrix);
  TVarRec v[] = {P.X, P.Y, P.Z};
  TextFile->Add(Format(" Start point of Text: X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  TextFile->Add(" The text is: " + S1);
  TextFile->Add(" Angle=" + FloatToStr(vText->Rotation + FCADParams.Angle));
}
//---------------------------------------------------------------------------

void __fastcall TForm1::ImportSpline(TObject *Sender)
{
  int I;
  TFPoint P;
  TsgDXFVertex *Vertex;
  TsgDXFSpline *vSpline = (TsgDXFSpline *)Sender;
  TVarRec v[3];

  TextFile->Add("Points of Spline:");

  int Cnt = vSpline->Count;
  for(I = 0; I < Cnt; I++)
  {
    Vertex = (TsgDXFVertex *)(vSpline->Fit->Items[I]);
    P = PtXMat(Vertex->Point, FCADParams.Matrix);
    v[0] = P.X;
    v[1] = P.Y;
    v[2] = P.Z;
    TextFile->Add(" P" + IntToStr(I + 1) + ": " +
      Format(" X=%.6f Y=%.6f Z=%.6f", v, ARRAYSIZE(v)-1));
  }
}
//---------------------------------------------------------------------------





