//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "fMain.h"
#include "fLayers.h"
#include "fAddScaledFile.h"
#include "Math.h"

#include <DXFConv.hpp>
#include <sgConsts.hpp>
#include <DXFConv.hpp>
#include <DXFImage.hpp>
#ifndef sgDXFONLY
#include <DWG.hpp>
#include <HPGL2.hpp>
#include <SVG.hpp>
#include <CGM.hpp>
#endif
#include <sgExport.hpp>
#include <ShellAPI.hpp>
#include <Printers.hpp>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "SGImage"
#pragma resource "*.dfm"
TfmMain *fmMain;
TsgDXFImage *Img;
int Start;
//---------------------------------------------------------------------------
__fastcall TfmMain::TfmMain(TComponent* Owner)
  : TForm(Owner)
{
}
//---------------------------------------------------------------------------

void _fastcall TfmMain::Time()
{
  int T;
  T = (GetCurrentTime() - Start) / 1000;
  TVarRec v[] = {T / 60, T % 60};
  Panel3->Caption.Format("%.2d:%.2d", v, 1);
  Panel3->Update();
}

void TfmMain::AddTree(TTreeNode * Parent, TsgDXFEntity * E)
{
    Parent = TreeView1->Items->AddChildObject(Parent, E->EntName(), E);
    Application->ProcessMessages();
    ProgressBar1->Position =+ 1;
    for (int i = 0; i < E->Count; i++) AddTree(Parent, E->Entities[i]);
}


void _fastcall TfmMain::LoadStructure()
{
  AnsiString S_LStructure;
  ProgressBar1->Position = 0;
  ProgressBar1->Max = Img->Converter->Count();
  Panel1->Caption = S_LStructure;
  Panel1->Update();
  TreeView1->Items->BeginUpdate();
  try
  {
    for (int i = 0; i < Img->Converter->Main->Count; i++)
      AddTree(NULL, Img->Converter->Main->Entities[i]);
  }
  __finally
  {
    ProgressBar1->Position = 0;
    TreeView1->Items->EndUpdate();
  }
}

TsgDXFImage * _fastcall TfmMain::GetImg(TGraphic *AGraphic)
{
  TsgDXFImage *vImg = NULL;
  if (AGraphic)
  {
    if (AGraphic->ClassType() == __classid(TsgDXFImage))
      vImg = dynamic_cast<TsgDXFImage *>(AGraphic);
    #ifndef sgDXFONLY
    if (AGraphic->ClassType() == __classid(TsgHPGLImage))
      vImg = dynamic_cast<TsgHPGLImage *>(AGraphic);
    if (AGraphic->ClassType() == __classid(TsgSVGImage))
      vImg = dynamic_cast<TsgSVGImage *>(AGraphic);
    if (AGraphic->ClassType() == __classid(TsgCGMImage))
      vImg = dynamic_cast<TsgCGMImage *>(AGraphic);
    #endif
  }
  return vImg;
}

void _fastcall TfmMain::EnableScales()
{
  MIPrint->Enabled = (!(sgPaintBox->Empty())&&(Img));
  miSaveAs->Enabled = (!(sgPaintBox->Empty())&&(Img));
  miSaveToDXF->Enabled = (!(sgPaintBox->Empty())&&(Img));
  sb3DOrbit->Enabled = (!(sgPaintBox->Empty())&&(Img));
  btnLayers->Enabled = (!(sgPaintBox->Empty())&&(Img));
  btnFitToScreen->Enabled = (!(sgPaintBox->Empty())&&(Img));
  MiOEM->Enabled = (!(sgPaintBox->Empty())&&(!Img));
  MIClose->Enabled = !(sgPaintBox->Empty());
  BtnSave->Enabled = !(sgPaintBox->Empty());
  BtnPrint->Enabled = !(sgPaintBox->Empty());
  BtnPrint1->Enabled = !(sgPaintBox->Empty());
  BtnPlus->Enabled = !(sgPaintBox->Empty());
  BtnMinus->Enabled = !(sgPaintBox->Empty());
  BtnReal->Enabled = !(sgPaintBox->Empty());
  BtnPrev->Enabled = !(sgPaintBox->Empty());
}

void _fastcall TfmMain::Open(AnsiString FileName)
{
  TreeView1->Items->Clear();
  Orbit3D->Visible = false;
  Start = GetCurrentTime();
  Screen->Cursor = crHourGlass;
  try
  {
    sgPaintBox->LoadFromFile(FileName);
    if (sgPaintBox->Picture->Graphic->InheritsFrom(__classid(Graphics::TBitmap)))
      sgPaintBox->Picture->Bitmap->HandleType = bmDDB;
    MIClose->Enabled = true;
    btnLayers->Enabled = true;
    Panel1->Caption = UpperCase(ExtractFileName(FileName));
    EnableScales();
    Img = GetImg(sgPaintBox->Picture->Graphic);
    if (!Img) return;
    HelpContext = 3;
    sb3DOrbit->Enabled = false;
    if (cbShowLineWeight->Checked)
      Img->ColorToLineWeight->Text = memColorToLineWeight->Text;
    Img->IsShowLineWeight = cbShowLineWeight->Checked;

    TabSheet2->TabVisible = true;
    Panel1->Caption = UpperCase(ExtractFileName(FileName));
    MIPrint->Enabled = true;
    MiOEM->Enabled = true;
    MiOEM->Checked = !(Img->Converter->OEM());
    MiOEMClick(NULL);
    sb3DOrbit->Enabled = true;
    sb3DOrbit->Down = false;
  }
  __finally
  {
    ProgressBar1->Position = 0;
    Screen->Cursor = crDefault;
    miSaveAs->Enabled = (Img != NULL);
    miSaveToDXF->Enabled = miSaveAs->Enabled;
    sgPaintBox->Align = alNone;
    sgPaintBox->Align = alClient;
  }
}

void __fastcall TfmMain::FormCreate(TObject *Sender)
{
  OpenDialog->Filter = GraphicFilter(__classid(TGraphic));
  /*if (ParamCount > 0)
  {
    Open(ParamStr(1));
    OpenDialog->FileName = ParamStr(1);
  } */
  Orbit3D = new TsgOrbit3D(this);
  Orbit3D->Parent = ScrollBox1;
  Orbit3D->Visible = false;
  Orbit3D->Drawing = sgPaintBox;
}

void TfmMain::Scroll(TShiftState Shift, bool Both, WORD Msg, WORD Code1,
  WORD Code2)
{
  if (Shift.Contains(ssCtrl)) Code1 = Code2;
  SendMessage(ScrollBox1->Handle, Msg, Code1, 0);
  if ((Both) && (Shift.Contains(ssCtrl)))
    SendMessage(ScrollBox1->Handle, WM_VSCROLL, Code1, 0);
}

void __fastcall TfmMain::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
  sgPaintBox->Cancel();
  if (Key == VK_ESCAPE) Close();
  if (sgPaintBox->Empty()) return;
  if ((PageControl1->ActivePage != TabSheet1) || (!Shift.Contains(ssCtrl))) return;
  switch (Key){
    case VK_ADD:
      BtnPlusClick(NULL);
      break;
    case VK_SUBTRACT:
      BtnMinusClick(NULL);
      break;
    case VK_MULTIPLY:
      BtnRealClick(BtnReal);
      break;
    case VK_UP:
      Scroll(Shift, false, WM_VSCROLL, SB_LINELEFT, SB_PAGELEFT);
      break;
    case VK_DOWN:
      Scroll(Shift, false, WM_VSCROLL, SB_LINERIGHT, SB_PAGERIGHT);
      break;
    case VK_PRIOR:
      Scroll(Shift, false, WM_VSCROLL, SB_PAGELEFT, SB_TOP);
      break;
    case VK_NEXT:
      Scroll(Shift, false, WM_VSCROLL, SB_PAGERIGHT, SB_BOTTOM);
      break;
    case VK_LEFT:
      Scroll(Shift, false, WM_HSCROLL, SB_LINELEFT, SB_PAGELEFT);
      break;
    case VK_RIGHT:
      Scroll(Shift, false, WM_HSCROLL, SB_LINERIGHT, SB_PAGERIGHT);
      break;
    case VK_HOME:
      Scroll(Shift, true, WM_HSCROLL, SB_TOP, SB_TOP);
      break;
    case VK_END:
      Scroll(Shift, true, WM_HSCROLL, SB_BOTTOM, SB_BOTTOM);
      break;
  }
}

void __fastcall TfmMain::miSaveAsClick(TObject *Sender)
{
  char *vFileName = StrNew("");

  if ((sgPaintBox->Empty()) || (!SavePictureDialog->Execute())) return;
  if (LowerCase(ExtractFileExt(SavePictureDialog->FileName)) == ".wmf")
  {
    TMetafile *vM = new TMetafile;
    try
    {
      vM->Width = sgPaintBox->Width;
      vM->Height = sgPaintBox->Height;
      TMetafileCanvas *vMC = new TMetafileCanvas(vM, 0);
      try
      {
        vMC->StretchDraw(sgPaintBox->ClientRect, sgPaintBox->Picture->Graphic);
      }
      __finally
      {
        delete vMC;
      }
      vM->SaveToFile(SavePictureDialog->FileName);
    }
    __finally
    {
      delete vM;
    }
    return;
  }

  TsgExport *E = new TsgExport;
  try
  {
    vFileName = "abviewer.tmp";
    vFileName = StrPCopy(vFileName, ExtractFileName(SavePictureDialog->FileName));
    if (ExtractFileExt(SavePictureDialog->FileName) != ".bmp")
      vFileName = StrCat(vFileName, ".bmp");
    E->ExportTo(sgPaintBox->Picture->Graphic, vFileName,
      Sgexport::TsgExportFormat(SavePictureDialog->FilterIndex - 1), sgPaintBox->Width,
      sgPaintBox->Height);
  }
  __finally
  {
    delete E;
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::OpenDialogShow(TObject *Sender)
{
  MinDXFSize = 140;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::OpenDialogClose(TObject *Sender)
{
  MinDXFSize = 0;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::miSaveToDXFClick(TObject *Sender)
{
  #ifdef SG_USEDXFEXPORT
  TsgDXFExport *E;
  TMetafile *vMet;
  TGraphic *vGr;
  TsgFloat vOffX, vOffY, vUSize;
  AnsiString vSaveToFile;

  if (!(SavePictureDialogDXF->Execute())) return;
  if (AnsiLowerCase(ExtractFileExt(SavePictureDialogDXF->FileName)) != ".dxf")
     SavePictureDialogDXF->FileName = SavePictureDialogDXF->FileName + ".dxf";
  vSaveToFile = SavePictureDialogDXF->FileName;
  vGr = (TGraphic *)(sgPaintBox->Picture->Graphic);
  E = new TsgDXFExport;
  try
  {
    E->IsParseWhite = true;// false
    if (!((vGr->ClassType() == __classid(TsgDXFImage)) || (vGr->ClassType() == __classid(TMetafile))))
    {
      ShowMessage("Raster drawings can not be saved to DXF format.");
      return;
    }
    else
    {
      if (vGr->ClassType() == __classid(TsgDXFImage))
      {
        E->Use01MM = false;
        E->Canvas->StretchDraw(Rect(0, 0, vGr->Width, vGr->Height), vGr);
        E->EndDraw();
      }
      else
      {
        vMet = ((TsgDXFImage *)(vGr))->ExportToMetafile(vOffX, vOffY, vUSize);
        OffsetX = vOffX;
        OffsetY = vOffY;
        E->UnitSize = vUSize;
        E->LoadFromMetafile(vMet);
        delete vMet;
      }
    }
    E->SaveToFile(vSaveToFile);
  }
  __finally
  {
    delete E;
  }
  #else
    MessageBox(this->Handle, "This function is available with the DXFExport VCL only.')", "WARNING", MB_ICONWARNING);
  #endif
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::MIExitClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::MICloseClick(TObject *Sender)
{
  ListBox1->Clear();
  TreeView1->Items->Clear();
  sgPaintBox->Picture->Graphic = NULL;
  Img = NULL;
  Panel1->Caption = "";
  Panel3->Caption = "";
  MIClose->Enabled = false;
  MIPrint->Enabled = false;
  MiOEM->Enabled = false;
  BtnSave->Enabled = false;
  PageControl1->ActivePage = TabSheet1;
  TabSheet2->TabVisible = false;
  sgPaintBox->Align = alNone;
  sgPaintBox->SetBounds(0,0,0,0);
  ProgressBar1->Max = 100;
  HelpContext = 1;
  EnableScales();
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::MiOEMClick(TObject *Sender)
{
  MiOEM->Checked = !MiOEM->Checked;
  if (Img == NULL) return;
  if (MiOEM->Checked)
    Img->Charset = OEM_CHARSET;
  else
    Img->Charset = DEFAULT_CHARSET;
  ListBox1->Font->Charset = Img->Charset;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::TreeView1Change(TObject *Sender, TTreeNode *Node)
//---------------------------------------------------------------------------
{
  if (!Node->Selected) return;
  TsgDXFEntity *E = ((TsgDXFEntity *)Node->Data);
  AnsiString S;
  if (E == NULL)
    ListBox1->Clear();
  else
  {
    if ((E->SrcStart == NULL) || (E->SrcEnd == NULL))
      ListBox1->Clear();
    else
    {
      S = AnsiString(E->SrcStart, E->SrcEnd - E->SrcStart);
      ListBox1->Items->Text = S;
    }
  }
}
//---------------------------------------------------------------------------
void __fastcall TfmMain::MIPrintClick(TObject *Sender)
{
  if (sgPaintBox->Width > sgPaintBox->Height)
    Printer()->Orientation = poLandscape;
  else
    Printer()->Orientation = poPortrait;
  if (PrintDialog1->Execute())
  {
    double W = Printer()->PageWidth / sgPaintBox->Picture->Width;
    double H = Printer()->PageHeight / sgPaintBox->Picture->Height;
    if (W > H) W = H;
    int PW = W * sgPaintBox->Picture->Width;
    int PH = W * sgPaintBox->Picture->Height;
    Printer()->BeginDoc();
    Printer()->Canvas->StretchDraw(Bounds((Printer()->PageWidth - PW) / 2,
      (Printer()->PageHeight - PH) / 2, PW, PH), sgPaintBox->Picture->Graphic);
    Printer()->EndDoc();
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::BtnPrevClick(TObject *Sender)
{
  sgPaintBox->PrevScale();
  btnFitToScreen->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sgPaintBoxProgress(TObject *Sender,
      TProgressStage Stage, BYTE PercentDone, bool RedrawNow,
      const TRect &R, const AnsiString Msg)
{
  if (Stage == psStarting)
  {
    Panel1->Caption = Msg;
    Panel1->Update();
  }
  ProgressBar1->Position = PercentDone;
  Time();
  if (Stage == psEnding)
  {
    ProgressBar1->Position = 0;
    Panel1->Caption = UpperCase(ExtractFileName(OpenDialog->FileName));
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::BtnPlusClick(TObject *Sender)
{
  TPoint p = sgPaintBox->Center();
  sgPaintBox->ChangeScale(false, 2, p);
  btnFitToScreen->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::BtnMinusClick(TObject *Sender)
{
  TPoint p = sgPaintBox->Center();
  sgPaintBox->ChangeScale(false, 0.5, p);
  btnFitToScreen->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::PageControl1Changing(TObject *Sender,
      bool &AllowChange)
{
  AllowChange = (Img != NULL);
  FocusControl(NULL);  
}
//---------------------------------------------------------------------------


void __fastcall TfmMain::ScrollBox1Resize(TObject *Sender)
{
  if (!(sgPaintBox->Empty()))
    sgPaintBox->SetBounds(sgPaintBox->Left,sgPaintBox->Top,
       sgPaintBox->Width,sgPaintBox->Height);  
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::BtnSaveClick(TObject *Sender)
{
  char *vFileName = StrNew("");

  if ((sgPaintBox->Empty()) || (!SavePictureDialog->Execute())) return;
  if (LowerCase(ExtractFileExt(SavePictureDialog->FileName)) == ".wmf")
  {
    TMetafile *vM = new TMetafile;
    try
    {
      vM->Width = sgPaintBox->Width;
      vM->Height = sgPaintBox->Height;
      TMetafileCanvas *vMC = new TMetafileCanvas(vM, 0);
      try
      {
        vMC->StretchDraw(sgPaintBox->ClientRect, sgPaintBox->Picture->Graphic);
      }
      __finally
      {
        delete vMC;
      }
      vM->SaveToFile(SavePictureDialog->FileName);
    }
    __finally
    {
      delete vM;
    }
    return;
  }

  TsgExport *E = new TsgExport;
  try
  {
    vFileName = "abviewer.tmp";
    vFileName = StrPCopy(vFileName, ExtractFileName(SavePictureDialog->FileName));
    if (ExtractFileExt(SavePictureDialog->FileName) != ".bmp")
      vFileName = StrCat(vFileName, ".bmp");
    E->ExportTo(sgPaintBox->Picture->Graphic, vFileName,
      Sgexport::TsgExportFormat(SavePictureDialog->FilterIndex - 1), sgPaintBox->Width,
      sgPaintBox->Height);
  }
  __finally
  {
    delete E;
  }
}
//---------------------------------------------------------------------------


void __fastcall TfmMain::Disabled1Click(TObject *Sender)
{
  if ((sgPaintBox->Empty()) || (Img == NULL)) return;
  ((TsgDXFImage *)sgPaintBox->Picture->Graphic)->DrawMode = dmGray;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::Black1Click(TObject *Sender)
{
  if ((sgPaintBox->Empty()) || (Img == NULL)) return;
  ((TsgDXFImage *)sgPaintBox->Picture->Graphic)->DrawMode = dmBlack;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::Normal1Click(TObject *Sender)
{
  if ((sgPaintBox->Empty()) || (Img == NULL)) return;
  ((TsgDXFImage *)sgPaintBox->Picture->Graphic)->DrawMode = dmNormal;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::FormMouseWheelDown(TObject *Sender,
      TShiftState Shift, TPoint &MousePos, bool &Handled)
{
  if ((sgPaintBox->Empty()) || (PageControl1->ActivePage != TabSheet1)) return;
  MousePos = ScrollBox1->ScreenToClient(MousePos);
  sgPaintBox->ChangeScale(false, 0.8, MousePos);
  Handled = true;
}
//---------------------------------------------------------------------------
void __fastcall TfmMain::FormMouseWheelUp(TObject *Sender,
      TShiftState Shift, TPoint &MousePos, bool &Handled)
{
  if ((sgPaintBox->Empty()) || (PageControl1->ActivePage != TabSheet1)) return;
  MousePos = ScrollBox1->ScreenToClient(MousePos);
  sgPaintBox->ChangeScale(false, 1.2, MousePos);
  Handled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::SoftGoldHomePage1Click(TObject *Sender)
{
  ShellExecute(Handle, "open", "www.cadsofttools.com", "", "", SW_SHOW);  
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::MIAboutClick(TObject *Sender)
{
  ShowMessage("DWG/DXF VCLs demo version \nPlease contact www.cadsofttools.com to order a full version");  
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::btnLayersClick(TObject *Sender)
{
  if ((Img == NULL)||(!(Img->Converter))) return;
  try
  {
    LayersDialogExecute(Img, sgPaintBox);
  }
  catch (Exception &exception)
  {
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sgPaintBoxPaint(TObject *Sender)
{
  if (Orbit3D->Visible)
  {
    Orbit3D->Paint();
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sb3DOrbitClick(TObject *Sender)
{
  sb3DOrbit->Down = sb3DOrbit->Down | false;
  if (sb3DOrbit->Down)
  {
    Orbit3D->CADImage = Img;
    Orbit3D->Visible = true;
    Orbit3D->Parent->Repaint();
  }
  else
  {
    Orbit3D->CADImage = NULL;
    Orbit3D->Visible = false;
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sgPaintBoxMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
  RECT vRect;
  TFPoint vPoint;
  TsgDXFImage *vImg;

  vImg = GetImg(sgPaintBox->Picture->Graphic);
  if (vImg)
  {
    vRect = sgPaintBox->PictureRect();
    vImg = dynamic_cast<TsgDXFImage *>(sgPaintBox->Picture->Graphic);
    vPoint = vImg->GetCADCoords((X * 1.0) / (vRect.right - vRect.left),
      1.0 - (Y * 1.0 / (vRect.bottom - vRect.top)));
    #ifndef sgDXFONLY
    if (sgPaintBox->Picture->Graphic->ClassType() == __classid(TsgHPGLImage))
    {
        TsgHPGLImage *vHPGLImg;
        vHPGLImg = dynamic_cast<TsgHPGLImage *>(sgPaintBox->Picture->Graphic);
        vPoint.X = vPoint.X * vHPGLImg->MMWidth / 100 / (vHPGLImg->Extents.Right - vHPGLImg->Extents.Left);
        vPoint.Y = vPoint.Y * vHPGLImg->MMHeight / 100 / ceil(vHPGLImg->Extents.Top - vHPGLImg->Extents.Bottom);
        vPoint.Z = 0.0;
    }
    #endif

    TVarRec v[] = {vPoint.X, vPoint.Y, vPoint.Z};
    pnlCoords->Caption = Format("(%.3f; %.3f; %.3f)", v, 2);
  }
  else
    pnlCoords->Caption = "";
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sbAddScaledFilesClick(TObject *Sender)
{
  TsgDXFImage *vCAD;
  Orbit3D->Visible = false;
  fmAddScaledFile = new TfmAddScaledFile(this);
  fmAddScaledFile->ShowModal();
  if ((fmAddScaledFile->CADFile != NULL)&&(fmAddScaledFile->FilesCount != 0))
  {
    vCAD = fmAddScaledFile->CADFile;
    btnLayers->Enabled = true;
    if (cbShowLineWeight->Checked)
      vCAD->ColorToLineWeight->Text = Memo1->Text;
    else
      vCAD->ColorToLineWeight->Text =  "";
    sgPaintBox->Picture->Graphic = vCAD;
    Img = GetImg(sgPaintBox->Picture->Graphic);
    if (Img)
    {
      sgPaintBox->Width = Ceil(Img->AbsWidth());
      sgPaintBox->Height = Ceil(Img->AbsHeight());
    }
    EnableScales();
    sb3DOrbit->Enabled = true;
    sb3DOrbit->Down = false;
    miSaveAs->Enabled = true;
    miSaveToDXF->Enabled = miSaveAs->Enabled;
    btnFitToScreenClick(Sender);
  }
  delete fmAddScaledFile;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::BtnPrintClick(TObject *Sender)
{
  if (sgPaintBox->Width > sgPaintBox->Height)
    Printer()->Orientation = poLandscape;
  else
    Printer()->Orientation = poPortrait;
  if (PrintDialog1->Execute())
  {
    double W = Printer()->PageWidth / sgPaintBox->Picture->Width;
    double H = Printer()->PageHeight / sgPaintBox->Picture->Height;
    if (W > H) W = H;
    int PW = W * sgPaintBox->Picture->Width;
    int PH = W * sgPaintBox->Picture->Height;
    Printer()->BeginDoc();
    Printer()->Canvas->StretchDraw(Bounds((Printer()->PageWidth - PW) / 2,
      (Printer()->PageHeight - PH) / 2, PW, PH), sgPaintBox->Picture->Graphic);
    Printer()->EndDoc();
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::BtnPrint1Click(TObject *Sender)
{
  double koef = 25.4;
  double PxlsX, PxlsY;
  int W, H;
  HDC hPrinter;

  if (sgPaintBox->Width > sgPaintBox->Height)
    Printer()->Orientation = poLandscape;
  else
    Printer()->Orientation = poPortrait;

  if (PrintDialog1->Execute())
  {
    W = Img->Width;
    H = Img->Height;
    PxlsX = 1;
    PxlsY = 1;
    if (sgPaintBox->Picture->Graphic->ClassType() == __classid(TsgDXFImage))
    {
      hPrinter = Printer()->Handle;
      PxlsX = GetDeviceCaps(hPrinter, LOGPIXELSX);
      PxlsY = GetDeviceCaps(hPrinter, LOGPIXELSY);
      if (Img->Millimetres)
      {
        PxlsX = PxlsX / koef;
        PxlsY = PxlsY / koef;
      }
    }
    Printer()->Title = "CAD Image VCL printing 1:1";
    Printer()->BeginDoc();
    Printer()->Canvas->StretchDraw(Rect(0, 0, Ceil(PxlsX * W), Ceil(PxlsY * H)), Img);
    Printer()->EndDoc();
  }
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::FormResize(TObject *Sender)
{
  if ((Orbit3D)&&(Orbit3D->Visible))
    Orbit3D->Refresh();
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::FormDestroy(TObject *Sender)
{
  delete Orbit3D;
  Orbit3D = NULL;
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::cbShowLineWeightClick(TObject *Sender)
{
  if (!(Img)) return;
  Img->IsShowLineWeight = cbShowLineWeight->Checked;
  if (cbShowLineWeight->Checked)
    Img->ColorToLineWeight->Text = memColorToLineWeight->Text;
  else
    Img->ColorToLineWeight->Text = "";
  sgPaintBox->Repaint();
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::PageControl1Change(TObject *Sender)
{
  if (!TreeView1->Items->Count) LoadStructure();
}
//---------------------------------------------------------------------------

void __fastcall TfmMain::sbMergingTestClick(TObject *Sender)
{
  TsgDXFImage *vGlobalCADFile;
  TsgDXFImage *Im;
  AnsiString cnstLayerName = "new_layer";
  AnsiString cnstLTypeName = "HIDDEN";
  AnsiString cnstCurDirName = "MergeFiles\\";
  char *cnstFileNames[2] = {"Frame.dxf", "Inserted.dxf"};
  char *N[8] = {"TABLES", "BLOCKS", "ENTITIES", "LTYPE", "LAYERS", "BLOCK_RECORD", "STYLE", "DIM_STYLE"};


  TObject *vCADFiles[2];
  TsgDXFLineType *vLType;
  TsgDXFLine *vLine;
  TsgDXFArc *vArc;
  TsgDXFTable *vTbl;
  int I;

  SetCurrentDir(ExtractFilePath(Application->ExeName) + cnstCurDirName);
  MacroStrings->Text = Memo1->Text;
  vGlobalCADFile = new TsgDXFImage;
  // Drawing have not border
  vGlobalCADFile->IsWithoutBorder = true;
  vGlobalCADFile->ColorToLineWeight->Add("clYellow = 0.1");
  vGlobalCADFile->ColorToLineWeight->Add("clLime = 0.3");
  vGlobalCADFile->ColorToLineWeight->Add("clAqua = 0.1");
  vGlobalCADFile->ColorToLineWeight->Add("clBlue= 0.1");
  vGlobalCADFile->ColorToLineWeight->Add("clFuchsia = 0.1");
  vGlobalCADFile->ColorToLineWeight->Add("clBlack = 0.1");
  vGlobalCADFile->ColorToLineWeight->Add("clGray= 0.5");
  vGlobalCADFile->ColorToLineWeight->Add("clSilver= 1");
  // or this: vGlobalCADFile->ColorToLineWeight->Assign(Self.memColorToLineWeight.Lines);
  // Adding the new linetype to image-converter
  vLType = new TsgDXFLineType;
  vLType->Name = cnstLTypeName;
  vLType->Lines->Scale = 50;        // linetype scale
  vLType->Lines->AddTick(0.25);      // length of the "solid" part of a line
  vLType->Lines->AddTick(-0.125);    // length of the "empty" part of a line
  if (vGlobalCADFile->Converter->Sections[csLTypes] == NULL)
  {
    vGlobalCADFile->Converter->Sections[csLTypes] = (TsgDXFGroup *)(vGlobalCADFile->Converter->NewEntity(vGlobalCADFile->Converter->Sections[csTables],__classid(TsgDXFTable)));
    vGlobalCADFile->Converter->Sections[csLayers] = (TsgDXFGroup *)(vGlobalCADFile->Converter->NewEntity(vGlobalCADFile->Converter->Sections[csTables],__classid(TsgDXFTable)));
    vGlobalCADFile->Converter->Main->AddEntity(vGlobalCADFile->Converter->Sections[csTables]);
    vGlobalCADFile->Converter->Main->AddEntity(vGlobalCADFile->Converter->Sections[csBlocks]);
    vGlobalCADFile->Converter->Main->AddEntity(vGlobalCADFile->Converter->Sections[csEntities]);
    for(I=0;I<4;I++)
    {
      vTbl = (TsgDXFTable *)(vGlobalCADFile->Converter->Sections[(TConvSection)I]);
      vTbl->Name = N[I];
    }
  }
  vGlobalCADFile->Converter->Sections[csLTypes]->AddEntity(vLType);

  // The first inserted file
  if (ExtractFileExt(cnstFileNames[0]) == ".dxf")
    vCADFiles[0] = new TsgDXFImage;
  #ifdef sgDXFONLY
  ;
  #else
  else
    vCADFiles[0] = new TsgDWGImage;
  #endif
  Im = dynamic_cast<TsgDXFImage *>(vCADFiles[0]);
  Im->LoadFromFile((AnsiString)(cnstFileNames[0]));
  vGlobalCADFile->AddScaledDXF(
    vCADFiles[0],                 // TsgDXFImage object for inserting
    ChangeFileExt(ExtractFileName(cnstFileNames[0]), ""), // Name
    MakeFPoint(0.0, 0.0, 0.0),    // Position in ACAD coords
    MakeFPoint(1, 1, 1),          // Scale factor
    0.0);                         // Rotation angle (degrees)

  // The second inserted file
  if (ExtractFileExt(cnstFileNames[1]) == ".dxf")
    vCADFiles[1] = new TsgDXFImage;
  #ifdef sgDXFONLY
  ;
  #else
  else
    vCADFiles[1] = new TsgDWGImage;
  #endif
  Im = dynamic_cast<TsgDXFImage *>(vCADFiles[1]);
  Im->LoadFromFile((AnsiString)(cnstFileNames[1]));
  vGlobalCADFile->AddScaledDXF(
    vCADFiles[1],                  // TsgDXFImage object for inserting
    ChangeFileExt(ExtractFileName(cnstFileNames[1]), ""), // Name
    MakeFPoint(80.0, 100.0, 0.0),  // position in ACAD coords
    MakeFPoint(10, 10, 1),         // scale factor
    0.0);                          // rotation angle (degrees)

  // Adding the line entity
  vLine = new TsgDXFLine;
  vLine->Point  = MakeFPoint( 80, 100, 0);// first point
  vLine->Point1 = MakeFPoint(150, 150, 0);// second point
  vLine->SetColor(clBlue);                 // color
  vLine->Layer = vGlobalCADFile->Converter->LayerByName(cnstLayerName);
  vLine->SetLType(vGlobalCADFile->Converter->LTypeByName(cnstLTypeName));
  if (vGlobalCADFile->Converter->OnCreate != NULL)
    vGlobalCADFile->Converter->OnCreate(vLine);
  vGlobalCADFile->Converter->Loads(vLine);
  vGlobalCADFile->Converter->Sections[csEntities]->AddEntity(vLine);

  // Adding the arc entity
  vArc = new TsgDXFArc;
  vArc->StartAngle = 60;                  // start angle in degrees
  vArc->EndAngle = 210;                   // } angle in degrees
  vArc->Radius = 50;                      // radius
  vArc->Point = MakeFPoint( 110, 160, 0); // center point
  vArc->SetColor(clBlue);                  // color
  vArc->Layer = vGlobalCADFile->Converter->LayerByName(cnstLayerName);
  vArc->SetLType(vGlobalCADFile->Converter->LTypeByName(cnstLTypeName));
  if (vGlobalCADFile->Converter->OnCreate != NULL)
    vGlobalCADFile->Converter->OnCreate(vArc);
  vGlobalCADFile->Converter->Loads(vArc);
  vGlobalCADFile->Converter->Sections[csEntities]->AddEntity(vArc);

  // Apply new drawing extents
  vGlobalCADFile->GetExtents();

  // Copying of new drawing to sgPaintBox
  btnLayers->Enabled = true;
  vGlobalCADFile->IsWithoutBorder = false;
  sgPaintBox->Picture->Graphic = vGlobalCADFile;
  Img = GetImg(sgPaintBox->Picture->Graphic);
  EnableScales();
  if (Img)
  {
    sgPaintBox->Width = Ceil(Img->AbsWidth());
    sgPaintBox->Height = Ceil(Img->AbsHeight());
  }
  sgPaintBox->Invalidate();
  Orbit3D->Visible = false;
  Orbit3D->CADImage = Img;
  sb3DOrbit->Enabled = true;
  sb3DOrbit->Down = false;
  sgPaintBox->Align = alNone;
  sgPaintBox->Align = alClient;
  delete vGlobalCADFile;
}

//---------------------------------------------------------------------------

void __fastcall TfmMain::btnLTypeHalfClick(TObject *Sender)
{
  int I, J;
  if (sgPaintBox->Empty()) return;
  if (Img)
  {
    if (Img == NULL) return;
    Img->Converter->LScale = Img->Converter->LScale / 2;
    for(I=0;I<Img->Converter->Counts[csEntities] - 1;I++)
      Img->Converter->Loads(Img->Converter->Entities[I]);
    for(I=0;I<Img->Converter->Counts[csBlocks] - 1;I++)
      for(J=0;J<Img->Converter->Blocks[I]->Count - 1;J++)
        Img->Converter->Loads(Img->Converter->Blocks[I]->Entities[J]);
    sgPaintBox->Invalidate();
  }
}
//---------------------------------------------------------------------------


void __fastcall TfmMain::BtnOpenClick(TObject *Sender)
{
  if (!OpenDialog->Execute()) return;
  MICloseClick(NULL);
  Open(OpenDialog->FileName);
}
//---------------------------------------------------------------------------


void __fastcall TfmMain::BtnRealClick(TObject *Sender)
{
  TPoint P;
  P = sgPaintBox->Center();
  sgPaintBox->ChangeScale(true, ((TComponent *)Sender)->Tag / 100.0, P);
  btnFitToScreen->Enabled = true;
}
//---------------------------------------------------------------------------



void __fastcall TfmMain::btnFitToScreenClick(TObject *Sender)
{
  sgPaintBox->Align = alNone;
  sgPaintBox->Align = alClient;
  btnFitToScreen->Enabled = false;
}
//---------------------------------------------------------------------------

void _fastcall TfmMain::sgPaintBoxScaling(TObject *Sender)
{
  sgPaintBox->Align = alNone;
  BtnReal->Enabled = true;
  btnFitToScreen->Enabled = true;
}
//---------------------------------------------------------------------------





  