unit Unit1;

//{$DEFINE SG_USEDXFEXPORT}// only for DXFExport VCL
//{$DEFINE SG_DEBUG}

{$IFDEF SG_DEBUG}
  {$DEFINE MEMCHK}
{$ENDIF}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, DXFImage, sgConsts,
  {$IFNDEF sgDXFONLY}
  DWG, HPGL2, CGM, SVG,
  {$ENDIF}
  {$IFDEF SG_USEDXFEXPORT}
  DXFExport,
  {$ENDIF}
  {$IFDEF MEMCHK}
  MemCheck,
  {$ENDIF}
  SGImage, ExtDlgs, DXFConv, ComCtrls;

const
  sNotVectorial: string = 'Raster drawings can not be saved to DXF format';

type
  TfmMain = class(TForm)
    pnlTop: TPanel;
    btnOpen: TButton;
    sbImage: TScrollBox;
    sgCADImage: TsgImage;
    OpenPictureDialog1: TOpenPictureDialog;
    btnZoomIn: TButton;
    btnZoomOut: TButton;
    btnNew: TButton;
    btnDelete: TButton;
    btnSaveToDXF: TButton;
    SavePictureDialog: TSavePictureDialog;
    Panel1: TPanel;
    sbrCoords: TStatusBar;
    procedure btnOpenClick(Sender: TObject);
    procedure btnZoomInClick(Sender: TObject);
    procedure btnZoomOutClick(Sender: TObject);
    procedure btnNewClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure sgCADImageMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure btnSaveToDXFClick(Sender: TObject);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
  private
    FLastAdded: TsgDXFEntity;
  public
    { Public declarations }
  end;

var
  fmMain: TfmMain;
  FSaveToFile: string = '';

implementation

uses fEntity;

{$R *.DFM}

procedure TfmMain.btnOpenClick(Sender: TObject);
begin
  if OpenPictureDialog1.Execute then
  begin
    sbrCoords.Panels.Items[0].Text := 'Please wait...';
    sgCADImage.LoadFromFile(OpenPictureDialog1.FileName);
    sgCADImage.Align := alNone;
    sgCADImage.Align := alClient;
  end;
  sbrCoords.Panels.Items[0].Text := OpenPictureDialog1.FileName;
end;

procedure TfmMain.btnZoomInClick(Sender: TObject);
begin
  sgCADImage.ChangeScale(False, 2, sgCADImage.Center);
end;

procedure TfmMain.btnZoomOutClick(Sender: TObject);
begin
  sgCADImage.ChangeScale(False, 0.5, sgCADImage.Center);
end;

procedure TfmMain.btnNewClick(Sender: TObject);
var
  Ent: TsgDXFEntity;
  Im: TsgDXFImage;
begin
  if not (sgCADImage.Picture.Graphic is TsgDXFImage) then
    Exit;
  Im := TsgDXFImage(sgCADImage.Picture.Graphic);
  Ent := GetEntity;
  if Ent <> nil then
  begin
    Im.Converter.Sections[csEntities].AddEntity(Ent);
    if Assigned(Im.Converter.OnCreate) then
      Im.Converter.OnCreate(Ent);
    Im.Converter.Loads(Ent);
    Im.GetExtents;
    sgCADImage.Refresh;
    FLastAdded := Ent;
  end;
end;

procedure TfmMain.FormCreate(Sender: TObject);
begin
  {$IFNDEF VER100}
  Self.OnMouseWheelDown := FormMouseWheelDown;
  Self.OnMouseWheelUp := FormMouseWheelUp;
  {$ENDIF}
  FLastAdded := nil;
end;

procedure TfmMain.btnDeleteClick(Sender: TObject);
var
  Im: TsgDXFImage;
begin
  if not (sgCADImage.Picture.Graphic is TsgDXFImage) then
    Exit;
  Im := TsgDXFImage(sgCADImage.Picture.Graphic);
  if FLastAdded <> nil then
  begin
    Im.Converter.DeleteEntity(FLastAdded, True);
    Im.GetExtents;
    FLastAdded := nil;
    sgCADImage.Refresh;
  end;
end;

procedure TfmMain.sgCADImageMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  vPt: TFPoint;

  function DoCADPosition(PointOnScr: TPoint): TFPoint;
  var
    vImg: TsgDXFImage;
    {$IFNDEF sgDXFONLY}
    vHPGLImg: TsgHPGLImage;
    {$ENDIF}
    vRect: TRect;
  begin
    vImg := TsgDXFImage(sgCADImage.Picture.Graphic);
    vRect := sgCADImage.PictureRect;
    Result := vImg.GetCADCoords(PointOnScr.X / (vRect.Right - vRect.Left),
       PointOnScr.Y / (vRect.Bottom - vRect.Top));
    {$IFNDEF sgDXFONLY}
    if sgCADImage.Picture.Graphic is TsgHPGLImage then
    begin
      vHPGLImg := TsgHPGLImage(sgCADImage.Picture.Graphic);
      Result.X := Result.X * vHPGLImg.MMWidth / 100 / (vHPGLImg.Extents.Right - vHPGLImg.Extents.Left);
      Result.Y := Result.Y * vHPGLImg.MMHeight / 100 / Abs(vHPGLImg.Extents.Top - vHPGLImg.Extents.Bottom);
    end;
    {$ENDIF}
  end;

begin
  if sgCADImage.Empty then Exit;
  if sgCADImage.Picture.Graphic is TsgDXFImage then
  begin
    vPt := DoCADPosition(Point(X, sgCADImage.Height - Y));
    sbrCoords.Panels.Items[1].Text := Format('(%.3f; %.3f; 0.000)', [vPt.X, vPt.Y]);
  end;
end;

procedure TfmMain.btnSaveToDXFClick(Sender: TObject);
{$IFDEF SG_USEDXFEXPORT}
var
  E: TsgDXFExport;
  vMet: TMetafile;
  vGr: TGraphic;
  vOffX, vOffY, vUSize: TsgFloat;
{$ENDIF}
begin
  {$IFDEF SG_USEDXFEXPORT}
  if not SavePictureDialog.Execute then Exit;
  if AnsiLowerCase(ExtractFileExt(SavePictureDialog.FileName)) <> '.dxf' then
     SavePictureDialog.FileName := SavePictureDialog.FileName + '.dxf';
  FSaveToFile := SavePictureDialog.FileName;
  vGr := sgCADImage.Picture.Graphic;
  E := TsgDXFExport.Create;
  try
    E.IsParseWhite := True;// False
    if not ((vGr is TsgDXFImage) or (vGr is TMetafile)) then
    begin
      ShowMessage(sNotVectorial);
      Exit;
    end
    else
    begin
      if not (vGr is TsgDXFImage) then
      begin
        E.Use01MM := False;
        E.Canvas.StretchDraw(Rect(0, 0, vGr.Width, vGr.Height), vGr);
        E.EndDraw;
      end
      else
      begin
        vMet := TsgDXFImage(vGr).ExportToMetafile(vOffX, vOffY, vUSize);
        DXFExport.OffsetX := vOffX;
        DXFExport.OffsetY := vOffY;
        E.UnitSize := vUSize;
        E.LoadFromMetafile(vMet);
        {$IFDEF SG_DEBUG}
        vMet.SaveToFile(ChangeFileExt(FSaveToFile, '.emf'));
        {$ENDIF}
        vMet.Free;
      end;
    end;
    E.SaveToFile(FSaveToFile);
  finally
    E.Free;
  end;
  {$ELSE}
  MessageBox(0, PChar('This function is available with the DXFExport VCL only.'), 'WARNING', MB_ICONWARNING);
  {$ENDIF}
end;

procedure TfmMain.FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
  if sgCADImage.Empty then Exit;
  MousePos := sbImage.ScreenToClient(MousePos);
  sgCADImage.ChangeScale(False, 0.8, MousePos);
  Handled := True;
end;

procedure TfmMain.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
  if sgCADImage.Empty then Exit;
  MousePos := sbImage.ScreenToClient(MousePos);
  sgCADImage.ChangeScale(False, 1.2, MousePos);
  if sgCADImage.Width < 5 then
    sgCADImage.ChangeScale(False, 2, MousePos);
  Handled := True;
end;

initialization
  {$IFDEF MEMCHK}
  MemChk
  {$ENDIF}

end.
