unit fPreview;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, SGImage, sgSManager, DXFImage, DXFConv;

type
  TfmPreview = class(TForm)
    ScrollBox1: TScrollBox;
    sgImage: TsgImage;
    Label1: TLabel;
    btnCancel: TButton;
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
      MousePos: TPoint; var Handled: Boolean);
    procedure sgImageMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    FSManager: TsgSelectionManager;
    FCurPos: TPoint;
  public
    constructor Create(AOwner: TComponent); override;
    procedure FillImage(ASManager: TsgSelectionManager; ADXFImage: TsgDXFImage);
  end;

  procedure CreatePreview(AOwner: TComponent; ASManager: TsgSelectionManager;
    ADXFImage: TsgDXFImage);

implementation
{$INCLUDE SGDXF.INC}

{$R *.dfm}

procedure CreatePreview(AOwner: TComponent; ASManager: TsgSelectionManager;
  ADXFImage: TsgDXFImage);
var
  Preview: TfmPreview;
begin
  if ASManager.SelectedEntities.Count < 1 then
  begin
    ShowMessage('There are no selected entities');
    Exit;
  end;
  Preview := TfmPreview.Create(AOwner);
  try
    Preview.FillImage(ASManager, ADXFImage);
    Preview.ShowModal;
  finally
    Preview.Free;
  end;
end;

{ TfmPrintPreview }

constructor TfmPreview.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCurPos := sgImage.Center;
end;

procedure TfmPreview.FillImage(ASManager: TsgSelectionManager;
  ADXFImage: TsgDXFImage);
var
  I: Integer;
  vEnt: TsgDXFEntity;
  vPoint: TPoint;
  vDXFImage: TsgDXFImage;
begin
  FSManager := ASManager;
  vDXFImage := TsgDXFImage.Create;
  for I := 0 to FSManager.SelectedEntities.Count - 1 do
  begin
    vEnt := TsgDXFEntityClass(TsgDXFEntity(FSManager.SelectedEntities[I]).ClassType).Create;
    vEnt.AssignEntity(TsgDXFEntity(FSManager.SelectedEntities[I]));
    vDXFImage.Converter.OnCreate(vEnt);
    vDXFImage.Converter.Sections[csEntities].AddEntity(vEnt);
    vDXFImage.Converter.Loads(vEnt);
  end;
  vDXFImage.GetExtents;
  sgImage.Picture.Graphic := vDXFImage;
  //TsgDXFImage(sgImage.Picture.Graphic).IsWithoutBorder := True;// without border
  vDXFImage.Free;
  vPoint.X := sgImage.Width div 2;
  vPoint.Y := sgImage.Height div 2;
  sgImage.ChangeScale(False, 1, vPoint);
end;

procedure TfmPreview.FormMouseWheelDown(Sender: TObject;
  Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin
  if sgImage.Width > 5 then
    sgImage.ChangeScale(False, 0.8, FCurPos);
  Handled := True;
end;

procedure TfmPreview.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
  MousePos: TPoint; var Handled: Boolean);
begin
  sgImage.ChangeScale(False, 1.2, FCurPos);
  Handled := True;
end;

procedure TfmPreview.sgImageMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  FCurPos := Point(X + sgImage.Left, Y + sgImage.Top);
end;

procedure TfmPreview.FormResize(Sender: TObject);
begin
  sgImage.ChangeScale(False, 1, FCurPos);
end;

procedure TfmPreview.FormCreate(Sender: TObject);
begin
  {$IFDEF SGDEL_4}
  OldCreateOrder := True;
  OnMouseWheelDown := FormMouseWheelDown;
  OnMouseWheelUp := FormMouseWheelUp;
  {$ENDIF}
  sgImage.Align := alNone;
  sgImage.Align := alClient;
end;

end.
