unit sgSManager;

interface

uses Windows, Messages, Classes, Controls, sgConsts, sgImage,
     DXFImage, DXFConv, Graphics, SysUtils, dialogs, Forms;

const
  BoxSize = 4;
  CursorRad = 5;
  SM_SELECTIONCHANGED = WM_USER + 2000;
type
  TsgManagerMode = (mmSelect, mmMultiSelect);

  TsgSelectionManager = class(TObject)
  private
    FsgImage: TsgImage;
    FImg: TsgDXFImage;
    FConverter: TsgDXFConverter;
    FDownPoint: TPoint;
    FCurPoint: TPoint;
    FSelection: Boolean;
    FMode: TsgManagerMode;
    FEntities: TList;
    FSChanged: Boolean;
    FDraging: Boolean;
    FStartPoint: TPoint;
    FOnMouseDownOld: TMouseEvent;
    FOnMouseMoveOld: TMouseMoveEvent;
    FOnMouseUpOld: TMouseEvent;
    FOnImagePaint: TNotifyEvent;
    procedure SetsgImage(const Value: TsgImage);
    procedure DrawSelection;
    procedure ClearSelection;
    procedure AddEntities(NewEntities: TList; Shift: TShiftState);
    function FindEntities(ARect: TRect): TList;
    function IsPtInBox(Point: TPoint; Entity: TsgDXFEntity): Boolean;
    function IsEntityInRect(Entity: TsgDXFEntity; Rect: TRect): Boolean;
    function IsPtOnTheLine(APoint, AStart, AnEnd: TPoint): Boolean;
    function GetRealEntityRect(AnEntity: TsgDXFEntity): TRect;
    function InList(AnEntity: TsgDXFEntity): Integer;
    function CheckList(NewEntities: TList; Shift: TShiftState): Boolean;
    function CheckListEnt(NewEntity: TsgDXFEntity; Shift: TShiftState): Boolean;
  protected
    procedure sgImageMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure sgImageMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure sgImageMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure sgImagePaint(Sender: TObject);
    function FindParentEntity(Point: TPoint): TsgDXFEntity;
 public
    constructor Create;
    destructor Destroy; override;
    function FindEntity(Point: TPoint): TsgDXFEntity;
    procedure Cancel;
    procedure SelectAll;
    procedure DeselectAll;
    property Entities: TList read FEntities;
    property sgImage: TsgImage read FsgImage write SetsgImage;
    property Mode: TsgManagerMode read FMode write FMode;
    property SelectedEntities: TList read FEntities;
  end;

procedure DrawFrame(AHDC: HDC; const ARect: TRect);

implementation

uses fEntityInspector, fMain;

procedure DrawFrame(AHDC: HDC; const ARect: TRect);
var
  Width, Height: Integer;
begin
  Width := ARect.Right - ARect.Left;
  Height := ARect.Bottom - ARect.Top;
  PatBlt(AHDC, ARect.Left, ARect.Top, Width, 1, PATINVERT);
  PatBlt(AHDC, ARect.Left + Width - 1, ARect.Top + 1, 1,
    Height - 2, PATINVERT);
  PatBlt(AHDC, ARect.Left, ARect.Top + Height - 1, Width, 1,
    PATINVERT);
  PatBlt(AHDC, ARect.Left, ARect.Top + 1, 1, Height - 2, PATINVERT);
end;

function TsgSelectionManager.FindEntity(Point: TPoint): TsgDXFEntity;
var
  I, J: Integer;
  vBox: TFRect;
  CADPoint: TFPoint;
begin
  Result := nil;
  for I := FConverter.Counts[csEntities] - 1 downto 0 do
  begin
    if FConverter.Entities[I] is TsgDXFInsert then// INSERT
    begin
      vBox := TsgDXFInsert(FConverter.Entities[I]).Box;
      CADPoint := FImg.GetCADCoords(Point.X / FsgImage.Width, 1 - Point.Y / FsgImage.Height);
      if (CADPoint.X > vBox.Left) and (CADPoint.X < vBox.Right) and
         (CADPoint.Y > vBox.Bottom) and (CADPoint.Y < vBox.Top) then
        Result := TsgDXFInsert(FConverter.Entities[I]);
    end
    else
      if IsPtInBox(Point, FConverter.Entities[I]) then
      begin
        if (FConverter.Entities[I] is TsgDXFText) or (FConverter.Entities[I] is TsgDXFMText) then
        begin
          Result := FConverter.Entities[I];
          Break;
        end;
        if FConverter.Entities[I] is TsgDXFLine then
        begin
          if IsPtOnTheLine(Point, Fimg.GetPoint(TsgDXFLine(FConverter.Entities[I]).Point),
             Fimg.GetPoint(TsgDXFLine(FConverter.Entities[I]).Point1)) then
          begin
            Result := FConverter.Entities[I];
            Break;
          end;
        end;
        if FConverter.Entities[I] is TsgDXFPolyLine then
        begin
          for J := 0 to TsgDXFPolyLine(FConverter.Entities[I]).PolyPoints.Count - 2 do
          begin
            if IsPtOnTheLine(Point, FImg.GetPoint(TsgDXFPolyLine(FConverter.Entities[I]).Points[J]),
               FImg.GetPoint(TsgDXFPolyLine(FConverter.Entities[I]).Points[J+1])) then
            begin
              Result := FConverter.Entities[I];
              Break;
            end;
          end;
        end;
      end;
      if Result <> nil then Break;
  end;
end;

function TsgSelectionManager.FindEntities(ARect: TRect): TList;
  procedure ChangeRectBounds(var ARect: TRect);
  var
    Temp: Integer;
  begin
    if ARect.Right < ARect.Left then
    begin
      Temp := ARect.Left;
      ARect.Left := ARect.Right;
      ARect.Right := Temp;
    end;
    if ARect.Bottom < ARect.Top then
    begin
      Temp := ARect.Top;
      ARect.Top := ARect.Bottom;
      ARect.Bottom := Temp;
    end;
  end;
var
  I: Integer;
begin
  ChangeRectBounds(ARect);
  Result := TList.Create;
  for I := FConverter.Counts[csEntities] - 1 downto 0 do
    if IsEntityInRect(FConverter.Entities[I], ARect) then
      if (FConverter.Entities[I] is TsgDXFLine) or (FConverter.Entities[I] is TsgDXFPolyLine) or
        (FConverter.Entities[I] is TsgDXFText) or (FConverter.Entities[I] is TsgDXFInsert) then
        Result.Add(FConverter.Entities[I]);
end;

function TsgSelectionManager.IsPtInBox(Point: TPoint;
  Entity: TsgDXFEntity): Boolean;
var
  vR: TRect;
begin
  Result := False;
  if (FsgImage = nil) or (Entity = nil) or IsBadRect(Entity.Box) then Exit;
  vR.TopLeft := FImg.GetPoint(Entity.Box.TopLeft);
  vR.BottomRight := FImg.GetPoint(Entity.Box.BottomRight);
  InflateRect(vR, 1, 1);
  Result := PtInRect(vR, Point);
end;

function TsgSelectionManager.IsEntityInRect(Entity: TsgDXFEntity; Rect: TRect): Boolean;
var
  vTL, vBR: TPoint;
begin
  Result := False;
  if IsBadRect(Entity.Box) then
    Exit;
  vTL := FImg.GetPoint(Entity.Box.TopLeft);
  vBR := FImg.GetPoint(Entity.Box.BottomRight);
  Result :=  PtInRect(Rect, vTL) and PtInRect(Rect, vBR);
end;

function TsgSelectionManager.IsPtOnTheLine(APoint, AStart, AnEnd: TPoint): Boolean;
const
  Eps  = 100;
var
  Det, Distance: Double;
  I, J, Rad: Integer;
  CurPoint: TPoint;
begin
  Result := False;
  Rad := Round(CursorRad * FsgImage.Scale);
  I := - Rad;
  while (I <= Rad) do
  begin
    J := - Rad;
    while (J <= CursorRad) do
    begin
      CurPoint.X := APoint.X + I;
      CurPoint.Y := APoint.Y + J;
      if AStart.Y = AnEnd.Y then
      begin
        if AStart.X <= AnEnd.X then
          if (CurPoint.Y = AStart.Y) and (CurPoint.X >= AStart.X) and (CurPoint.X <= AnEnd.X) then
            Result := true;
        if AStart.Y <= AnEnd.Y then
          if (CurPoint.Y = AStart.Y) and (CurPoint.X > AStart.X) and (CurPoint.X < AnEnd.X) then
            Result := true;
      end
      else
      begin
        Det:= CurPoint.X * (AStart.Y - AnEnd.Y) - CurPoint.Y * (AStart.X - AnEnd.X) +
          AStart.X * AnEnd.Y - AnEnd.X * AStart.Y;
        Distance := (CurPoint.Y - AStart.Y) / (AnEnd.Y - AStart.Y);
        if (Abs(Det) <= eps) and (Abs(Distance) > 0) and (Abs(Distance) < 1) then
          Result := True;
      end;
      if Result then
        Break;
      J := J + 1;
    end;
    if Result then
      Break;
    I := I + 1;
  end;
end;

procedure TsgSelectionManager.SetsgImage(const Value: TsgImage);
begin
  if Assigned(Value) then
  begin
    if Assigned(FsgImage) and (FsgImage <> Value) then
    begin
      FsgImage.OnMouseDown := FOnMouseDownOld;
      FsgImage.OnMouseMove := FOnMouseMoveOld;
      FsgImage.OnMouseUp := FOnMouseUpOld;
      FsgImage.OnPaint := FOnImagePaint;
      FOnMouseDownOld := Value.OnMouseDown;
      FOnMouseMoveOld := Value.OnMouseMove;
      FOnMouseUpOld := Value.OnMouseUp;
    end
    else
      FsgImage := Value;
    if Assigned(FsgImage) then
    begin
      FsgImage.MouseRect := False;
      FsgImage.OnMouseDown := sgImageMouseDown;
      FsgImage.OnMouseMove := sgImageMouseMove;
      FsgImage.OnMouseUp := sgImageMouseUp;
      FsgImage.OnPaint := sgImagePaint;
      FImg := TsgDXFImage(FsgImage.Picture.Graphic);
      FConverter := FImg.Converter;
    end;
    if FsgImage.Picture.Graphic is TsgDXFImage then
      fmEntityInspector.Load(FsgImage);
  end
  else
    FsgImage := nil;
end;

procedure TsgSelectionManager.sgImageMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button <> mbLeft then Exit;
  FDownPoint := Point(X, Y);
  FCurPoint := FDownPoint;
  FSelection := True;
  if FMode = mmMultiSelect then
    DrawFrame(FsgImage.Canvas.Handle, Rect(FDownPoint.x, FDownPoint.y, FDownPoint.x, FDownPoint.y));
  FStartPoint.X := X;
  FStartPoint.Y := Y;
  if (FEntities.Count > 0) and (FMode <> mmMultiSelect) then
    FDraging := True;
  if Assigned(FOnMouseDownOld) then
    FOnMouseDownOld(Sender, Button, Shift, X, Y);
end;

procedure TsgSelectionManager.sgImageMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  vRect: TRect;
begin
  if FSelection then
  begin
    case FMode of
      mmMultiSelect:
      begin
        vRect.TopLeft := FDownPoint;
        vRect.BottomRight := FCurPoint;
        DrawFrame(FsgImage.Canvas.Handle, vRect);
        FCurPoint := Point(X, Y);
        vRect.BottomRight := FCurPoint;
        DrawFrame(FsgImage.Canvas.Handle, vRect);
      end;
    end;
  end;
  if Assigned(FOnMouseMoveOld) then
    FOnMouseMoveOld(Sender, Shift, X, Y);
end;

procedure TsgSelectionManager.sgImageMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  vRect: TRect;
  Pos: Integer;
  I, J: Integer;
  vMoveRight, vMoveTop: Real;
  FSelectedObject: TsgDXFEntity;
  vPoly: TsgDXFPolyline;
  vLine: TsgDXFLine;
  vSpline: TsgDXFSpline;
  vInvRect: TRect;

  function MovePoint(APoint: TFPoint): TFPoint;
  begin
    Result := MakeFPoint(APoint.X + vMoveRight, APoint.Y + vMoveTop, APoint.Z);
  end;
begin
  if Button <> mbLeft then Exit;
  FSelectedObject := nil;
  if (FStartPoint.X = X) and (FStartPoint.Y = Y) then
  begin
    ClearSelection;
    FSChanged := False;
    if FSelection then
    begin
      case FMode of
        mmSelect:
        begin
          FSChanged := CheckListEnt(FindEntity(Point(X, Y)), Shift);
          if not (ssShift in Shift) then
            FEntities.Clear;
          FSelectedObject := FindEntity(Point(X, Y));
          if FSelectedObject <> nil then
          begin
            Pos := InList(FSelectedObject);
            if Pos < 0 then
              FEntities.Add(FSelectedObject)
            else
            begin
              FSChanged := False;
              if ssShift in Shift then
              begin
                FEntities.Delete(Pos);
                FEntities.Capacity := FEntities.Capacity - 1;
                FSChanged := True;
              end;
            end;
          end
        end;
      end;
      DrawSelection;
      {if FSChanged then
        SendMessage(FsgImage.Parent.Parent.Handle, SM_SELECTIONCHANGED, 0, 0);}
      FSChanged := False;
      FSelection := False;
    end;
  end;
  if FMode = mmMultiSelect then
  begin
    ClearSelection;
    vRect.TopLeft := FDownPoint;
    vRect.BottomRight := FCurPoint;
    DrawFrame(FsgImage.Canvas.Handle, vRect);
    AddEntities(FindEntities(vRect), Shift);
    DrawSelection;
    FSelection := False;
  end;
  if FDraging then
  begin
    FDraging := False;
    vMoveRight := (X - FStartPoint.X) / FsgImage.Scale;
    vMoveTop := (FStartPoint.Y - Y) / FsgImage.Scale;

    for J := 0 to FEntities.Count - 1 do
    begin
      FSelectedObject := TsgDXFEntity(FEntities[J]);
      if FSelectedObject is TsgDXFSpline then
      begin
        vSpline := TsgDXFSpline(FSelectedObject);
        for I := 0 to vSpline.ControlCount - 1 do
          PFPoint(vSpline.Controls[I])^ := MovePoint(vSpline.ControlPoints[I]);
      end
      else
        if (FSelectedObject is TsgDXFEllipse) or (FSelectedObject is TsgDXFCircle) then
          TsgDXFEllipse(FSelectedObject).Point := MovePoint(TsgDXFEllipse(FSelectedObject).Point)
        else
          if FSelectedObject is TsgDXFPolyline then
          begin
            vPoly := TsgDXFPolyline(FSelectedObject);
            for I := 0 to vPoly.Count - 1 do
              TsgDXFVertex(vPoly.Entities[I]).Point := MovePoint(TsgDXFVertex(vPoly.Entities[I]).Point);
          end
          else
            if FSelectedObject is TsgDXFLine then
            begin
              vLine := TsgDXFLine(FSelectedObject);
              vLine.Point := MovePoint(vLine.Point);
              vLine.Point1 := MovePoint(vLine.Point1);
            end
            else
              if FSelectedObject is TsgDXFCustomVertex then
              begin// INSERT, DIMENSION, TEXT, MTEXT
                TsgDXFCustomVertex(FSelectedObject).Point := MovePoint(TsgDXFCustomVertex(FSelectedObject).Point);
                //if vText.HasSecond then// if TEXT
                //  vText.Point1 := MovePoint(vText.Point1);
              end;
      FConverter.Loads(FSelectedObject);
    end;
    vInvRect := fmMain.sgImage.PictureRect;
    InvalidateRect(FsgImage.Parent.Handle, @vInvRect, False);
    FSelection := False;
  end;

  if FEntities.Count>1 then FSelectedObject := nil;
  fmEntityInspector.RefreshGrid(FSelectedObject);

  if FEntities.Count = 0 then
    FSelection := False;
  if Assigned(FOnMouseUpOld) then
    FOnMouseUpOld(Sender, Button, Shift, X, Y);
end;

function TsgSelectionManager.FindParentEntity(Point: TPoint): TsgDXFEntity;
var
  I: Integer;
begin
  Result := nil;
  for I := FConverter.Counts[csEntities] - 1 downto 0 do
    if IsPtInBox(Point, FConverter.Entities[I]) and (FConverter.Entities[I] is TsgDXFPolyLine) then
    begin
      Result := FConverter.Entities[I];
      Break;
    end;
end;

function TsgSelectionManager.GetRealEntityRect(AnEntity: TsgDXFEntity): TRect;
begin
  Result.TopLeft := FImg.GetPoint(AnEntity.Box.TopLeft);
  Result.BottomRight := FImg.GetPoint(AnEntity.Box.BottomRight);
  Result.Left := FsgImage.Left + Result.Left - BoxSize;
  Result.Right := FsgImage.Left + Result.Right + 1 + BoxSize;
  Result.Top := FsgImage.Top + Result.Top - BoxSize;
  Result.Bottom := FsgImage.Top + Result.Bottom + 1 + BoxSize;
end;

procedure TsgSelectionManager.ClearSelection;
var
  I: Integer;
  vRect: TRect;
begin
  for I := 0 to FEntities.Count - 1 do
  begin
    vRect := GetRealEntityRect(TsgDXFEntity(FEntities[I]));
    InvalidateRect(FsgImage.Parent.Handle, @vRect, False);
  end;
end;

procedure TsgSelectionManager.DrawSelection;
var
  I, J: Integer;
  vRect: TRect;
  vLeft, vRight: TPoint;
  vBlockBorder: TRect;
  Cl: TColor;

  procedure DrawMarker(APoint: TPoint);
  begin
    vRect.Top := APoint.Y + BoxSize;
    vRect.Left := APoint.X - BoxSize;
    vRect.Bottom := APoint.Y - BoxSize;
    vRect.Right := APoint.X + BoxSize;
    FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
  end;
begin
  for I := 0 to FEntities.Count - 1 do
  begin
    if TsgDXFEntity(FEntities[I]) is TsgDXFLine then
    begin
      DrawMarker(FImg.GetPoint(TsgDXFLine(FEntities[I]).Point));
      DrawMarker(FImg.GetPoint(TsgDXFLine(FEntities[I]).Point1));
    end;
    if TsgDXFEntity(FEntities[I]) is TsgDXFPolyLine then
      for J := 0 to TsgDXFPolyLine(FEntities[I]).PolyPoints.Count - 1 do
        DrawMarker(FImg.GetPoint(TsgDXFPolyLine(FEntities[I]).Points[J]));
    if (TsgDXFEntity(FEntities[I]) is TsgDXFText) or (TsgDXFEntity(FEntities[I]) is TsgDXFMText) then
    begin
      vLeft := FImg.GetPoint(TsgDXFEntity(FEntities[I]).Box.TopLeft);
      vRight := FImg.GetPoint(TsgDXFEntity(FEntities[I]).Box.BottomRight);
      DrawMarker(vLeft);
      DrawMarker(vRight);
      vRect.Top := vRight.Y + BoxSize;
      vRect.Left := vLeft.X - BoxSize;
      vRect.Bottom := vRight.Y - BoxSize;
      vRect.Right := vLeft.X + BoxSize;
      FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
      vRect.Top := vLeft.Y + BoxSize;
      vRect.Bottom := vLeft.Y - BoxSize;
      vRect.Left := vRight.X - BoxSize;
      vRect.Right := vRight.X + BoxSize;
      FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
      vRect.Top := vLeft.Y + BoxSize;
      vRect.Bottom := vLeft.Y - BoxSize;
      vRect.Left := vLeft.X + (vRight.X - vLeft.X) div 2 - BoxSize;
      vRect.Right := vLeft.X + (vRight.X - vLeft.X) div 2 + BoxSize;
      FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
      vRect.Top := vRight.Y + BoxSize;
      vRect.Bottom := vRight.Y - BoxSize;
      vRect.Left := vLeft.X + (vRight.X - vLeft.X) div 2 - BoxSize;
      vRect.Right := vLeft.X + (vRight.X - vLeft.X) div 2 + BoxSize;
      FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
      vRect.Top := vRight.Y + (vLeft.Y - vRight.Y) div 2 + BoxSize;
      vRect.Bottom := vRight.Y + (vLeft.Y - vRight.Y) div 2 - BoxSize;
      vRect.Left := vLeft.X - BoxSize;
      vRect.Right := vLeft.X + BoxSize;
      FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
      vRect.Top := vRight.Y + (vLeft.Y - vRight.Y) div 2 + BoxSize;
      vRect.Bottom := vRight.Y + (vLeft.Y - vRight.Y) div 2 - BoxSize;
      vRect.Left := vRight.X - BoxSize;
      vRect.Right := vRight.X + BoxSize;
      FsgImage.Canvas.Rectangle(vRect.Left, vRect.Top, vRect.Right, vRect.Bottom);
    end
    else
      if TsgDXFEntity(FEntities[I]) is TsgDXFInsert then
      begin
        vBlockBorder.TopLeft := FImg.GetPoint(TsgDXFInsert(FEntities[I]).Box.TopLeft);
        vBlockBorder.BottomRight := FImg.GetPoint(TsgDXFInsert(FEntities[I]).Box.BottomRight);
        Cl := FsgImage.Canvas.Pen.Color;
        try
          FsgImage.Canvas.Pen.Color := clRed;
          FsgImage.Canvas.MoveTo(vBlockBorder.Left, vBlockBorder.Bottom);
          FsgImage.Canvas.LineTo(vBlockBorder.Right, vBlockBorder.Bottom);
          FsgImage.Canvas.LineTo(vBlockBorder.Right, vBlockBorder.Top);
          FsgImage.Canvas.LineTo(vBlockBorder.Left, vBlockBorder.Top);
          FsgImage.Canvas.LineTo(vBlockBorder.Left, vBlockBorder.Bottom);
        finally
          FsgImage.Canvas.Pen.Color := Cl;
        end;
      end;
  end;
end;

procedure TsgSelectionManager.AddEntities(NewEntities: TList; Shift: TShiftState);
var
  I: Integer;
begin
  FSChanged := CheckList(NewEntities, Shift);
  if not (ssShift in Shift) then
    FEntities.Clear;
  for I := 0 to NewEntities.Count - 1 do
    if InList(NewEntities[I]) < 0 then
      FEntities.Add(NewEntities[I]);
  NewEntities.Free;
end;

procedure TsgSelectionManager.sgImagePaint(Sender: TObject);
begin
  DrawSelection;
  if Assigned(FOnImagePaint) then
    FOnImagePaint(Sender);
end;

procedure TsgSelectionManager.Cancel;
begin
  ClearSelection;
  FEntities.Clear;
  DrawSelection;
  fmEntityInspector.Clear;
end;

procedure TsgSelectionManager.SelectAll;
var
  I: Integer;
begin
  ClearSelection;
  if FEntities.Count <> FConverter.Counts[csEntities] then
    FSChanged := True
  else
    FSChanged := False;
  for I := 0 to FConverter.Counts[csEntities] - 1 do
    if (FConverter.Entities[I] is TsgDXFLine) or (FConverter.Entities[I] is TsgDXFPolyLine) or
      (FConverter.Entities[I] is TsgDXFText) or (FConverter.Entities[I] is TsgDXFMText) then
      FEntities.Add(FConverter.Entities[I]);
  DrawSelection;
  {if FSChanged then
    SendMessage(FsgImage.Parent.Parent.Handle, SM_SELECTIONCHANGED, 0, 0);}
  FSChanged := False;
end;

procedure TsgSelectionManager.DeselectAll;
//var vBlockBorder: TFRect;
begin
  ClearSelection;
  if FEntities.Capacity <> 0 then
    FSChanged := True
  else
    FSChanged := False;
  FEntities.Clear;
  {vBlockBorder := FSelectedObject.Box;
  InvalidateRect(FsgImage.Parent.Handle, @vBlockBorder, False);
  if FSChanged then
    SendMessage(FsgImage.Parent.Parent.Handle, SM_SELECTIONCHANGED, 0, 0); }
  FSChanged := False;
  DrawSelection;
end;

function TsgSelectionManager.InList(AnEntity: TsgDXFEntity): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to FEntities.Count - 1 do
    if AnEntity = FEntities[I] then
    begin
      Result := I;
      Break;
    end;
end;

function TsgSelectionManager.CheckList(NewEntities: TList; Shift: TShiftState): Boolean;
var
  I: Integer;
begin
  Result := False;
  if NewEntities.Capacity = 0 then
  begin
    if (FEntities.Capacity <> 0) and (not(ssShift in Shift)) then
      Result := True;
  end
  else
  begin
    for I := 0 to NewEntities.Count - 1 do
      if InList(NewEntities[I]) < 0 then
        Result := True;
  end;
end;

function TsgSelectionManager.CheckListEnt(NewEntity: TsgDXFEntity; Shift: TShiftState): Boolean;
begin
  Result := False;
  if NewEntity = nil then
  begin
    if (FEntities.Capacity <> 0) and (not(ssShift in Shift)) then
      Result := True;
  end
  else
  begin
    if InList(NewEntity) < 0 then
        Result := True;
  end;
end;

constructor TsgSelectionManager.Create;
begin
  FEntities := TList.Create;
end;

destructor TsgSelectionManager.Destroy;
begin
  if sgImage <> nil then
  begin
    sgImage.OnMouseDown := FOnMouseDownOld;
    sgImage.OnMouseMove := FOnMouseMoveOld;
    sgImage.OnMouseUp := FOnMouseUpOld;
  end;
  FEntities.Free;
  inherited Destroy;
end;

end.
