unit Unit1;
{
  How to get entities properties.

  1. Create procedure of type:
    TsgCADEntityProc = procedure(Entity: TsgDXFEntity) of object;

    in this demo:
    procedure TForm1.ReadCADEntities(Entity: TsgDXFEntity);

   2. Crate TsgCADIterate object:
    FCADParams: TsgCADIterate; in this demo


   3. Create TsgDXFImage object (in the demo it is made in Button1Click
   on   Image1.Picture.LoadFromFile(OpenPictureDialog1.FileName);
   and stored in Image1.Picture.Graphic.

   4.  Set (TsgDXFImage object).Converer.AutoInsert := True
   See
   TsgDXFImage(Image1.Picture.Graphic).Converter.AutoInsert := True; // to get all the elements inside of inserts

   5. Call (TsgDXFImage object).Converer.Iterate with parameters made in items 1. and 2.
     TsgDXFImage(Image1.Picture.Graphic).Converter.Iterate(ReadCADEntities, FCADParams);

   6.  in procedure made in item 1. read all the entities:
       procedure TForm1.ReadCADEntities(Entity: TsgDXFEntity);
       begin
       //  use Entity here, see documentation about properties of different classes,
       //  allows to get lines, text, arcs, etc.
       end;
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtDlgs, ExtCtrls, DXFImage, DXFConv, Math, Extrusion,
  {$IFNDEF sgDXFONLY}
  DWG, HPGL2, SVG, CGM,
  {$ENDIF}
   sgConsts, MVFont;

type
  TForm1 = class(TForm)
    Image1: TImage;
    Panel1: TPanel;
    Button1: TButton;
    OpenPictureDialog1: TOpenPictureDialog;
    SaveDialog1: TSaveDialog;
    procedure Button1Click(Sender: TObject);
    procedure OpenPictureDialog1Show(Sender: TObject);
    procedure OpenPictureDialog1Close(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    TextFile: TStringList;
    FCADParams: TsgCADIterate;
    procedure ImportSolid(Sender: TObject);
    procedure ImportLine(Sender: TObject);
    procedure ImportEllipse(Sender: TObject);
    procedure ImportArc(Sender: TObject);
    procedure ImportCircle(Sender: TObject);
    procedure ImportPolyLine(Sender: TObject);
    procedure ImportAttdef(Sender: TObject);
    procedure ImportText(Sender: TObject);
    procedure ImportSpline(Sender: TObject);
    { Private declarations }
  public
    procedure ReadCADEntities(Entity: TsgDXFEntity);
    { Public declarations }
  end;

var
  Form1: TForm1;
  Img: TsgDXFImage;

implementation

{$R *.dfm}

{ All elements inside inserts can be reached in the next procedure.
  Proc called for Inserts too }
procedure TForm1.ReadCADEntities(Entity: TsgDXFEntity);
const
  PS: array[psSolid..psDashDot] of string =
    ('psSolid','psDash','psDot','psDashDot');
var
  S: string;
  L: TsgDXFLayer;
  C: TColor;
begin
  S := 'ClassName=' + Entity.ClassName + '; Entity name=' + Entity.EntName;
  DoScale2D(FCADParams); // calculates 2d scale and rotation
  // layer
  L := EntLayer(Entity, FCADParams.Insert);
  if L <> nil then S := S + ' layer = ' + L.Name;
  // color
  S := S + ' style = ' + PS[EntStyle(Entity)];
  C := EntColor(Entity, FCADParams.Insert);
  if C = clNone then
    S := S + ' color=black/white'
  else
    S := S + ' color = ' +  IntToHex(C, -6) + ' (' + ColorToString(C) + ')';
  // particular properties
  TextFile.Add(S);

  //viewport - see dxfimage.pas TsgDXFImage.Draw
  //hatch - see dxfimage.pas TsgDXFImage.DrawHatch
  //ellipse - see dxfimage.pas TsgDXFImage.DrawEllipse
  if Entity is TsgDXFEllipse then
  begin
    ImportEllipse(Entity);
    Exit;
  end;
  //arc - see dxfimage.pas TsgDXFImage.DrawArc
  if Entity is TsgDXFArc then
  begin
    ImportArc(Entity);
    Exit;
  end;
  //Circle - see dxfimage.pas TsgDXFImage.DrawCircle
  if Entity is TsgDXFCircle then
  begin
    ImportCircle(Entity);
    Exit;
  end;
  //Leader - see dxfimage.pas TsgDXFImage.DrawLeader
  //Spline - see dxfimage.pas TsgDXFImage.DrawSpline
  if Entity is TsgDXFSpline then
  begin
    ImportSpline(Entity);
    Exit;
  end;
  //PolyLine - see dxfimage.pas TsgDXFImage.DrawPolyLine
  if Entity is TsgDXFPolyLine then
  begin
    ImportPolyLine(Entity);
    Exit;
  end;
  //attdef - see dxfimage.pas TsgDXFImage.DrawText
  if Entity is TsgDXFAttdef then
  begin
    ImportAttdef(Entity);
    Exit;
  end;
  //text - see dxfimage.pas TsgDXFImage.DrawText
  if Entity is TsgDXFText then
  begin
    ImportText(Entity);
    Exit;
  end;
  //Face - see dxfimage.pas TsgDXFImage.DrawFace
  //Solid  - see dxfimage.pas TsgDXFImage.DrawSolid
  if Entity is TsgDXFSolid then
  begin
    ImportSolid(Entity);
    Exit;
  end;
  //point - see dxfimage.pas TsgDXFImage.DrawPoint
  //line  - see dxfimage.pas TsgDXFImage.DrawLine
  if Entity is TsgDXFLine then
  begin
    ImportLine(Entity);
    Exit;
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
const
  Exts: array[0..17] of string = ( '.dxf', '.dwg', '.plt', '.hgl', '.rtl',
    '.hg', '.plo', '.hp', '.hp1', '.hp2', '.hpg', '.hpgl', '.hpgl2', '.gl2',
    '.prn', '.spl', '.svg', '.cgm');
var
  vFileExt: string;
begin
  TextFile := TStringList.Create;
  try
    OpenPictureDialog1.FileName := '';
    if OpenPictureDialog1.Execute then
    begin
      vFileExt := ExtractFileExt(LowerCase(OpenPictureDialog1.FileName));
      case StrIndex(vFileExt, Exts) of
        0:	Img := TsgDXFImage.Create;
        {$IFNDEF sgDFONLY}
        1:	Img := TsgDWGImage.Create;
        2..15:	Img := TsgHPGLImage.Create;
        16:     Img := TsgSVGImage.Create;
        17:     Img := TsgCGMImage.Create;
        {$ENDIF}
        else	Img := nil;
      end;
      if Img <> nil then
      begin
        Img.LoadFromFile(OpenPictureDialog1.FileName);
        Image1.Picture.Assign(Img);
        FCADParams.Matrix := IdentityMat;
        TsgDXFImage(Image1.Picture.Graphic).Converter.ImportMode := imImport;
        TsgDXFImage(Image1.Picture.Graphic).Converter.AutoInsert := True; // to get all the elements inside of inserts
        TsgDXFImage(Image1.Picture.Graphic).Converter.Iterate(ReadCADEntities, FCADParams);
        SaveDialog1.FileName := ChangeFileExt(OpenPictureDialog1.FileName, '.txt');
        if SaveDialog1.Execute then
          TextFile.SaveToFile(SaveDialog1.FileName);
      end;
    end;
  finally
    TextFile.Free;
  end;
end;

procedure TForm1.OpenPictureDialog1Show(Sender: TObject);
begin
  {$IFNDEF sgDXFONLY}
  DWGPreview := True;
  {$ENDIF}
end;

procedure TForm1.OpenPictureDialog1Close(Sender: TObject);
begin
  {$IFNDEF sgDXFONLY}
  DWGPreview := False;
  {$ENDIF}
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  OpenPictureDialog1.Filter := GraphicFilter(TGraphic);
end;

procedure TForm1.ImportSolid(Sender: TObject);
var
  P: TFPoint;
  S: string;
begin
  P := PtXMat(TsgDXFSolid(Sender).Point, FCADParams.Matrix);
  S := S + ' P1: ';
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  P := PtXMat(TsgDXFSolid(Sender).Point1, FCADParams.Matrix);
  S := S + ' P2: ';
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  P := PtXMat(TsgDXFSolid(Sender).Point3, FCADParams.Matrix);
  S := S + ' P3: ';
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  P := PtXMat(TsgDXFSolid(Sender).Point2, FCADParams.Matrix);
  S := S + ' P4: ';
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  TextFile.Add(S);
end;

procedure TForm1.ImportLine(Sender: TObject);
var
  P: TFPoint;
  S: string;
begin
  S := ' Thickness=' + FloatToStr(TsgDXFLine(Sender).Thickness) + '; ';
  S := S + ' LineTypeScale=' + FloatToStr(TsgDXFLine(Sender).LineTypeScale) + '; ';
  S := S + ' ZThick=' + FloatToStr(TsgDXFLine(Sender).ZThick) + '; ';
  
  S := S + #13#10;
  P := PtXMat(TsgDXFLine(Sender).Point, FCADParams.Matrix);
  S := S + ' Begin point: ';
  S := S + #13#10;
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  P := PtXMat(TsgDXFLine(Sender).Point1, FCADParams.Matrix);
  S := S + ' End point: ';
  S := S + #13#10;
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  TextFile.Add(S);
end;

procedure TForm1.ImportEllipse(Sender: TObject);
var
  P: TFPoint;
  S: string;
begin
    P := PtXMat(TsgDXFEllipse(Sender).Point, FCADParams.Matrix);
    S := S + ' Center of Ellipse: ';
    S := S + ' X=' + FloatToStr(P.X);
    S := S + ' Y=' + FloatToStr(P.Y);
    S := S + ' Z=' + FloatToStr(P.Z);
    S := S + #13#10;
    S := S + ' Start Angle: ' + FloatToStr(TsgDXFEllipse(Sender).StartAngle);
    S := S + #13#10;
    S := S + ' End Angle: ' + FloatToStr(TsgDXFEllipse(Sender).EndAngle);
    S := S + #13#10;
    S := S + ' Rx: ' + FloatToStr(TsgDXFEllipse(Sender).Radius);
    S := S + ' Ry: ' + FloatToStr(TsgDXFEllipse(Sender).Radius * TsgDXFEllipse(Sender).Ratio);
    S := S + #13#10;
    TextFile.Add(S);
end;

procedure TForm1.ImportArc(Sender: TObject);
var
  P: TFPoint;
  S: string;
begin
  P := PtXMat(TsgDXFArc(Sender).Point, FCADParams.Matrix);
  S := S + ' Center of RoundArc: ';
  S := S + ' X=' + FloatToStr(P.X);
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  S := S + ' Start Angle: ' + FloatToStr(TsgDXFArc(Sender).StartAngle);
  S := S + #13#10;
  S := S + ' End Angle: ' + FloatToStr(TsgDXFArc(Sender).EndAngle);
  S := S + #13#10;
  S := S + ' Rx: ' + FloatToStr(TsgDXFArc(Sender).Radius);
  S := S + #13#10;
  TextFile.Add(S);
end;

procedure TForm1.ImportCircle(Sender: TObject);
var
  P: TFPoint;
  S: string;
begin
    P := PtXMat(TsgDXFCircle(Sender).Point, FCADParams.Matrix);
    S := S + ' Center of Circle: ';
    S := S + ' X=' + FloatToStr(P.X);
    S := S + ' Y=' + FloatToStr(P.Y);
    S := S + ' Z=' + FloatToStr(P.Z);
    S := S + #13#10;
    TextFile.Add(S);
end;
procedure TForm1.ImportPolyLine(Sender: TObject);
var
  P: TFPoint;
  S: string;
  I: Integer;
  Vertex: TsgDXFVertex;
begin
  if TsgDXFPolyLine(Sender).Closed
  then
    S := S + ' Points of Rect: '
  else
    S := S + ' Points of PolyLine: ';
  for I := 0 to TsgDXFPolyLine(Sender).Count - 1 do
  begin
    S := S + #13#10;
    S := S + ' P' + IntToStr(I + 1) + ': ';
    Vertex := TsgDXFVertex(TsgDXFPolyLine(Sender).Entities[I]);
    P := PtXMat(Vertex.Point, FCADParams.Matrix);
    S := S + ' X=' + FloatToStr(P.X);
    S := S + ' Y=' + FloatToStr(P.Y);
    S := S + ' Z=' + FloatToStr(P.Z);
  end;
  S := S + #13#10;
  TextFile.Add(S);
end;

function ImportTextFont(Sender: TObject): string;
var
  vText: TsgDXFText;

  function ImportMTextFontStyle(AStyle: TmvFontStyles): string;
  begin
    if fmBold in AStyle then Result := '  fmBold' + #13#10;
    if fmItalic in AStyle then Result := Result + '  fmItalic' + #13#10;
    if fmUnderline in AStyle then Result := Result + '  fmUnderline' + #13#10;
    if fmStrikeOut in AStyle then Result := Result + '  fmStrikeOut' + #13#10;
    if fmCondensed in AStyle then Result := Result + '  fmCondensed' + #13#10;
    if fmUpward in AStyle then Result := Result + '  fmUpward' + #13#10;
    if fmDownward in AStyle then Result := Result + '  fmDownward' + #13#10;
  end;

begin
  vText := TsgDXFText(Sender);
  if vText.Style <> nil then
  begin
    Result := Result + ' Text Style:' + #13#10;
    Result := Result + '  BigFont=' + vText.Style.BigFont + #13#10;
    Result := Result + '  WidthFactor=' + FloatToStr(vText.Style.WidthFactor) + #13#10;
  end;
  Result := ' Font: Name=' + vText.FontName + '; ';
  Result := Result + ' Height=' + IntToStr(vText.Font.Height) + '; ';
  Result := Result + ' Color=$' + IntToHex(vText.Font.Color, 8) + '; ' + #13#10;
  Result := Result + ' Style:' + #13#10 + ImportMTextFontStyle(vText.Font.Style);
  Result := Result + ' Thickness=' + FloatToStr(vText.GetThickness) + #13#10;
  Result := Result + ' Generation=' + IntToStr(vText.Generation) + #13#10;
  Result := Result + ' VAlign=' + IntToStr(vText.VAlign) + #13#10;
  Result := Result + ' HAlign=' + IntToStr(vText.HAlign) + #13#10;
end;

procedure TForm1.ImportAttdef(Sender: TObject);
var
  P: TFPoint;
  S, S1, S2: string;
  CurAttrib: TsgDXFAttrib;
  vText: TsgDXFText;
  vBox: TFRect;

begin
  DoScale2D(FCADParams);
  TsgDXFAttdef(Sender).DoNewText(TsgDXFAttdef(Sender).Tag);
  vText := TsgDXFText(Sender);
  S1 := vText.Text;
  P := PtXMat(vText.StartPoint, FCADParams.Matrix);
  S2 := ' Angle=' + FloatToStr(vText.Rotation);

  if FCADParams.Insert <> nil then
  begin
    CurAttrib := FCADParams.Insert.Attrib(S1, P);
    vText := CurAttrib;
    if CurAttrib <> nil then
    begin
      S1 := CurAttrib.Text;
      vBox := CurAttrib.Box;
      P := vBox.TopLeft;
      S2 := ' Angle=' + FloatToStr(CurAttrib.Rotation);
    end;
  end;
  S := S + ' Start point of Attdef: ';
  S := S + #13#10;
  S := S + ' X=' + FloatToStr(P.X);
  S := S + #13#10;
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + #13#10;
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  S := S + ' Text of Attdef: ' + S1;
  S := S + S2 + '; ';
  S := S + #13#10;
  if vText <> nil then
    S := S + ImportTextFont(vText);
  TextFile.Add(S);
end;

procedure TForm1.ImportText(Sender: TObject);
var
  P: TFPoint;
  S, S1: string;
begin
  DoScale2D(FCADParams);
  S1 := TsgDXFText(Sender).Text;
  P := PtXMat(TsgDXFText(Sender).StartPoint, FCADParams.Matrix);
  S := S + ' Start point of Text: ';
  S := S + #13#10;
  S := S + ' X=' + FloatToStr(P.X);
  S := S + #13#10;
  S := S + ' Y=' + FloatToStr(P.Y);
  S := S + #13#10;
  S := S + ' Z=' + FloatToStr(P.Z);
  S := S + #13#10;
  S := S + ' The text is: ' + S1;
  S := S + ' Angle=' + FloatToStr(TsgDXFText(Sender).Rotation + FCADParams.Angle);
  S := S + #13#10;
  S := S + ImportTextFont(Sender);
  TextFile.Add(S);
end;

procedure TForm1.ImportSpline(Sender: TObject);
var
  P: TFPoint;
  S: string;
  I: Integer;
  Vertex: TsgDXFVertex;
begin
    S := S + ' Points of Spline: ';
  for I := 0 to TsgDXFSpline(Sender).FitCount - 1 do
  begin
    S := S + #13#10;
    S := S + ' P' + IntToStr(I + 1) + ': ';
    Vertex := TsgDXFVertex(TsgDXFSpline(Sender).Fit[I]);
    P := PtXMat(Vertex.Point, FCADParams.Matrix);
    S := S + ' X=' + FloatToStr(P.X);
    S := S + ' Y=' + FloatToStr(P.Y);
    S := S + ' Z=' + FloatToStr(P.Z);
  end;
  S := S + #13#10;
  TextFile.Add(S);
end;

end.


