unit fLayers;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtDlgs, Buttons, StdCtrls, 
  ComCtrls, DXFImage, CheckLst, sgImage;

type
  TfmLayers = class(TForm)
    btnOK: TButton;
    btnCancel: TButton;
    ImageList: TImageList;
    clbLayers: TCheckListBox;
    procedure edtNameKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure clbLayersClickCheck(Sender: TObject);
    procedure clbLayersDrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure clbLayersKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    FControl: TControl;
    FImage: TsgDXFImage;
    { Private declarations }
  public
    { Public declarations }
  end;
var
  LayersRect: TRect;

function LayersDialogExecute(AImage: TsgDXFImage; Control: TControl): Boolean;

implementation

uses DXFConv;

{$R *.dfm}

function SetXrefLayerVisibility(AImage: TsgDXFImage; const AFullLayerName: string; AVisible: Boolean): Integer;
var
  I, vPos: Integer;
  vStrName, vStrLayerName: string;
begin
  vPos := Pos('|', AFullLayerName);
  if vPos <> 0 then
  begin
    vStrName := Copy(AFullLayerName, 1, vPos - 1);
    vStrLayerName := Copy(AFullLayerName, vPos + 1, Length(AFullLayerName) - vPos + 1);
    for I := 0 to AImage.Converter.XRefs.Count - 1 do
      if vStrName = TsgDXFXref(AImage.Converter.XRefs[I]).Name then
      begin
        TsgDXFImage(TsgDXFXref(AImage.Converter.XRefs[I]).CADImage).Converter.LayerByName(vStrLayerName).Visible := AVisible;
        Result := 1;
        Exit;
      end;
    for I := 0 to AImage.Converter.XRefs.Count - 1 do
    begin
      Result := SetXrefLayerVisibility(TsgDXFImage(TsgDXFXref(AImage.Converter.XRefs[I]).CADImage), AFullLayerName, AVisible);
      if Result = 1 then
        Exit;
    end
  end;
end;

function LayersDialogExecute(AImage: TsgDXFImage; Control: TControl): Boolean;
var
  I: Integer;
  vSLs: TList;
begin
  if AImage = nil then Exit;
  with TfmLayers.Create(nil) do
    try
      FControl := Control;
      //if LayersRect.Left <> 0 then
      //  BoundsRect := LayersRect;
      FImage := AImage;
      for I := 0 to AImage.Converter.Counts[csLayers] - 1 do
      begin
        clbLayers.Items.AddObject(AImage.Converter.Layers[I].Name, AImage.Converter.Layers[I]);
        clbLayers.Checked[I] := AImage.Converter.Layers[I].Visible;
      end;
      if clbLayers.Items.Count > 0 then
        clbLayers.ItemIndex := 0;
      vSLs := TList.Create;
      try
        for I := 0 to AImage.Converter.Counts[csLayers] - 1 do
          vSLs.Add(Pointer(AImage.Converter.Layers[I].Visible));
        ShowModal;
        Result := ModalResult = mrOk;
        if ModalResult = mrCancel then
        begin
          for I := 0 to AImage.Converter.Counts[csLayers] - 1 do
          begin
            AImage.Converter.Layers[I].Visible := Boolean(vSLs[I]);
            SetXrefLayerVisibility(AImage, AImage.Converter.Layers[I].Name, Boolean(vSLs[I]));
          end;
          FControl.Invalidate;
        end;
      finally
        vSLs.Free;
      end;
      LayersRect := BoundsRect;
    finally
      Free;
    end;
end;

procedure TfmLayers.edtNameKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
  begin
    Close;
    ModalResult := mrOk;
  end;
end;

procedure TfmLayers.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE: Close;
  end;
end;

procedure TfmLayers.clbLayersClickCheck(Sender: TObject);
var
  I: Integer;
  sgImage: TsgImage;
  vPt: TPoint;
begin
  for I := 0 to clbLayers.Items.Count - 1 do
    if TsgDXFLayer(clbLayers.Items.Objects[I]).Visible <> clbLayers.Checked[I] then
      TsgDXFLayer(clbLayers.Items.Objects[I]).Visible := clbLayers.Checked[I];
  FImage.GetExtents;
  if FControl is TsgImage then
  begin
    sgImage := TsgImage(FControl);
    vPt.X := sgImage.Width div 2;
    vPt.Y := sgImage.Height div 2;
    sgImage.ChangeScale(False, 1, vPt);
  end
  else
    FControl.Invalidate;
end;

procedure TfmLayers.clbLayersDrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
begin
  if Index > clbLayers.Items.Count - 1 then Exit;
  clbLayers.Canvas.Pen.Color := clWindow;
  clbLayers.Canvas.Rectangle(Rect.Left, Rect.Top, Rect.Right, Rect.Bottom);
  clbLayers.Canvas.Pen.Color := clWindowFrame;
  clbLayers.Canvas.Brush.Color := TsgDXFLayer(clbLayers.Items.Objects[Index]).Pen.Color;
  clbLayers.Canvas.Rectangle(Rect.Left, Rect.Top + 2, Rect.Left + 16, Rect.Bottom - 2);
  clbLayers.Canvas.Brush.Color := clWindow;
  if State = [] then
  begin
    clbLayers.Canvas.Brush.Color := clWindow;
    clbLayers.Canvas.Pen.Color := clWindowText;
  end
  else
  begin
    clbLayers.Canvas.Brush.Color := clHighlight;
    clbLayers.Canvas.Pen.Color := clHighlightText;
  end;
  clbLayers.Canvas.TextOut(Rect.Left + 24, Rect.Top, clbLayers.Items[Index]);
  clbLayers.Canvas.Brush.Color := clWindow;
  clbLayers.Canvas.Pen.Color := clWindowText;
end;

procedure TfmLayers.clbLayersKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
    ModalResult := mrOk;
end;

end.
