unit fLayers;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtDlgs, Buttons, StdCtrls, 
  DXFImage, ImgList, CheckLst, sgImage, ComCtrls;

type
  TfmLayers = class(TForm)
    btnOK: TButton;
    btnCancel: TButton;
    ImageList: TImageList;
    lvLayers: TListView;
    edSearch: TEdit;
    procedure edtNameKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure clbLayersKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormCreate(Sender: TObject);
    procedure lvLayersAdvancedCustomDrawSubItem(Sender: TCustomListView;
      Item: TListItem; SubItem: Integer; State: TCustomDrawState;
      Stage: TCustomDrawStage; var DefaultDraw: Boolean);
    procedure lvLayersResize(Sender: TObject);
    procedure lvLayersDblClick(Sender: TObject);
    procedure lvLayersDeletion(Sender: TObject; Item: TListItem);
  private
    FControl: TControl;
    FImage: TsgDXFImage;
    function GetColumnNumber(APoint: TPoint): Integer;
  public
    { Public declarations }
  end;

var
  LayersRect: TRect;

function LayersDialogExecute(AImage: TsgDXFImage; Control: TControl): Boolean;

implementation

uses DXFConv;

type
  PLayers = ^TLayers;
  TLayers = record
    Color: TColor;
    Layer: TsgDXFLayer;
    Frozen: Boolean;
    IsPlotting: Boolean;
    Visible: Boolean;
  end;

var
  SSearch: string = 'Search for layer';

{$R *.dfm}

function LayersDialogExecute(AImage: TsgDXFImage; Control: TControl): Boolean;
var
  I: Integer;
  vItem: TListItem;
  P: PLayers;
  vLayer: TsgDXFLayer;
begin
  Result := False;
  if AImage = nil then Exit;
  with TfmLayers.Create(nil) do
    try
      FControl := Control;
      BoundsRect := LayersRect;
      FImage := AImage;
      for I := 0 to AImage.Converter.Counts[csLayers] - 1 do
      begin
        vItem := lvLayers.Items.Add;
        vItem.Caption := AImage.Converter.Layers[I].Name;
        New(P);
        P^.Color := AImage.Converter.Layers[I].Color;
        P^.Layer := AImage.Converter.Layers[I];
        P^.Frozen := AImage.Converter.Layers[I].Frozen;
        P^.IsPlotting := AImage.Converter.Layers[I].IsPlotting;
        P^.Visible := AImage.Converter.Layers[I].Visible;
        vItem.Data := P;
      end;
      ShowModal;
      Result := ModalResult = mrOk;
      if ModalResult = mrCancel then
      begin
        for I := 0 to lvLayers.Items.Count - 1 do
        begin
          P := PLayers(lvLayers.Items[I].Data);
          vLayer := P^.Layer;
          vLayer.Color := P^.Color;
          vLayer.Frozen := P^.Frozen;
          vLayer.IsPlotting := P^.IsPlotting;
          vLayer.Visible := P^.Visible;
        end;
        FControl.Invalidate;
      end;
      LayersRect := BoundsRect;
    finally
      Free;
    end;
end;

procedure TfmLayers.edtNameKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
  begin
    Close;
    ModalResult := mrOk;
  end;
end;

procedure TfmLayers.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE: Close;
  end;
end;

procedure TfmLayers.clbLayersKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_RETURN then
    ModalResult := mrOk;
end;

procedure TfmLayers.FormCreate(Sender: TObject);
var
  I: Integer;
  An: TAnchors;
begin
  if not (PixelsPerInch = 96) then
  begin
    for I := 0 to ControlCount - 1 do
      if Controls[I] is TWinControl then
      begin
        An := TWinControl(Controls[I]).Anchors;
        TWinControl(Controls[I]).Anchors := [akLeft, akTop];
        TWinControl(Controls[I]).ScaleBy(96, PixelsPerInch);
        Controls[I].Left := Round(Controls[I].Left * 96 / PixelsPerInch);
        Controls[I].Top := Round(Controls[I].Top * 96 / PixelsPerInch);
        if PixelsPerInch = 120 then
        begin
          Controls[I].Top := Controls[I].Top - 8;
        end;
        TWinControl(Controls[I]).Anchors := An;
      end;
  end;
end;

procedure TfmLayers.lvLayersAdvancedCustomDrawSubItem(
  Sender: TCustomListView; Item: TListItem; SubItem: Integer;
  State: TCustomDrawState; Stage: TCustomDrawStage;
  var DefaultDraw: Boolean);
var
  vGraph: TGraphic;
  vLayer: TsgDXFLayer;
  vRect: TRect;
  I, X, Y: Integer;
begin
  vGraph := TIcon.Create;
  try
    vLayer := PLayers(Item.Data)^.Layer;
    vRect := Item.DisplayRect(drBounds);
    case SubItem of
      1:
        if vLayer.Visible then
          ImageList.GetIcon(0, TIcon(vGraph))
        else
          ImageList.GetIcon(2, TIcon(vGraph));
      2:
        if vLayer.Frozen then
          ImageList.GetIcon(5, TIcon(vGraph))
        else
          ImageList.GetIcon(6, TIcon(vGraph));
      3:
        begin
          vGraph.Free;
          vGraph := TBitmap.Create;
          vGraph.Width := 13;
          vGraph.Height := 13;
          TBitmap(vGraph).Canvas.Brush.Color := clWhite;
          TBitmap(vGraph).Canvas.FillRect(Rect(0, 0, 13, 13));
          TBitmap(vGraph).Canvas.Brush.Color := vLayer.Color;
          TBitmap(vGraph).Canvas.FillRect(Rect(0, 0, 12, 12));
          TBitmap(vGraph).Canvas.Pen.Width := 1;
          TBitmap(vGraph).Canvas.Pen.Color := clBlack;
          TBitmap(vGraph).Canvas.Brush.Style := bsClear;
          TBitmap(vGraph).Canvas.Rectangle(0, 0, 12, 12);
          TBitmap(vGraph).TransparentColor := clWhite;
        end;
      4:
        if vLayer.IsPlotting then
          ImageList.GetIcon(3, TIcon(vGraph))
        else
          ImageList.GetIcon(4, TIcon(vGraph));
    end;
    X := vRect.Left;
    Y := vRect.Top;
    for I := 0 to SubItem - 1 do
      Inc(X, Sender.Column[I].Width);
    if Item.Selected then
      Sender.Canvas.Brush.Color := clHighlight
    else
      Sender.Canvas.Brush.Color := clWindow;
    Sender.Canvas.FillRect(Rect(X, Y, X + Sender.Column[SubItem].Width,
      Y + Abs(vRect.Bottom - vRect.Top)));
    Inc(X, Sender.Column[SubItem].Width div 2);
    if SubItem in [1, 3] then
      Sender.Canvas.Draw(X - 7, Y, vGraph)
    else
    begin
      vRect := Rect(X - 7, Y, X + 7, Y + 14);
      Sender.Canvas.StretchDraw(vRect, vGraph);
    end;
  finally
    vGraph.Free;
  end;
  DefaultDraw := False;
end;

procedure TfmLayers.lvLayersResize(Sender: TObject);
begin
  Invalidate;
end;

procedure TfmLayers.lvLayersDblClick(Sender: TObject);
var
  vPt: TPoint;
  vIndex: Integer;
  vLayer: TsgDXFLayer;
  vColorDialog: TColorDialog;
  sgImage: TsgImage;
  vCurrLayout: TsgDXFLayout;
begin
  if GetCursorPos(vPt) then
  begin
    vPt := lvLayers.ScreenToClient(vPt);
    vIndex := GetColumnNumber(vPt);
    if vIndex <> -1 then
    begin
      vLayer := PLayers(lvLayers.Selected.Data)^.Layer;
      case vIndex of
        1: vLayer.Visible := not vLayer.Visible;
        2: vLayer.Frozen := not vLayer.Frozen;
        3:
          begin
            vColorDialog := TColorDialog.Create(Self);
            try
              vColorDialog.Options := [];
              vColorDialog.Color := vLayer.Color;
              if vColorDialog.Execute then
                vLayer.Color := vColorDialog.Color;
            finally
              vColorDialog.Free;
            end;
          end;
        4: vLayer.IsPlotting := not vLayer.IsPlotting;
      end;
      if vIndex <> 0 then
      begin
        vCurrLayout := FImage.CurrentLayout;
        FImage.GetExtents;
        FImage.CurrentLayout := vCurrLayout;
        if FControl is TsgImage then
        begin
          sgImage := TsgImage(FControl);
          vPt.X := sgImage.Width div 2;
          vPt.Y := sgImage.Height div 2;
          sgImage.ChangeScale(False, 1, vPt);
        end
        else
          FControl.Invalidate;
        lvLayers.Refresh;
      end;
    end;
  end;
end;

function TfmLayers.GetColumnNumber(APoint: TPoint): Integer;
var
  I, X: Integer;
  vRect, vBounds: TRect;
begin
  Result := -1;
  if lvLayers.Selected = nil then Exit;
  vBounds := lvLayers.Selected.DisplayRect(drBounds);
  X := vBounds.Left;
  vRect := Rect(X, vBounds.Top, X + lvLayers.Columns[0].Width, vBounds.Top +
    Abs(vBounds.Right - vBounds.Top));
  for I := 0 to lvLayers.Columns.Count - 1 do
  begin
    vRect.Left := X;
    vRect.Right := X + lvLayers.Columns[I].Width;
    Inc(X, lvLayers.Columns[I].Width);
    if PtInRect(vRect, APoint) then
    begin
      Result := I;
      Exit;
    end;
  end;
end;

procedure TfmLayers.lvLayersDeletion(Sender: TObject; Item: TListItem);
begin
  Dispose(Item.Data);
end;

end.
